#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "global.h"
#include "ali.h"
#include "macros.h"
#include "sprob.h"

#ifdef DEBUG
extern FILE *debugfile;
#endif

/*********************************************************
copy an alignment using deep-copy
**********************************************************/
ALI *copyAli (ALI *ali)
{
  int i;
  ALI *newali=NULL;
  size_t deslen;

  /* allocate memory for new alignment */
  MALLOC (newali, ALI, 1);

  /* copy elementary values */
  newali->id=ali->id;
  newali->num=ali->num;
  newali->len=ali->len;
  newali->score=ali->score;
  
  /* alloacte memory for pointers to descriptors, sequences, sids, structural information */
  MALLOC (newali->descriptors, char*, newali->num);
  MALLOC (newali->sequences, char*, newali->num);
  MALLOC (newali->sids, int, newali->num);
  MALLOC (newali->sinfo, SINFO*, newali->num);
  
  /* for every sequence in alignment */
  for (i=0; i<newali->num; i++)
    {
      /* descriptor */
      deslen=strlen (ali->descriptors[i]);
      MALLOC (newali->descriptors[i], char, deslen+1);
      strncpy (newali->descriptors[i], ali->descriptors[i], deslen+1);
      
      /* sequence itself */
      MALLOC (newali->sequences[i], char, newali->len+1);
      strncpy (newali->sequences[i], ali->sequences[i], newali->len+1);

      /* structural information */
      MALLOC (newali->sinfo[i], SINFO, newali->len);
      memcpy (newali->sinfo[i], ali->sinfo[i], (size_t)(newali->len)*sizeof (SINFO));
    }

  /* copy sequence ids */
  memcpy (newali->sids, ali->sids, (size_t)newali->num*sizeof (int));

  return (newali);
}

/*******************************************************
deletes a single column in the alignment
*******************************************************/
void deleteAliCol (ALI *ali, int col)
{
  int i;

  if (col>=ali->len) FATAL ("deleteAliCol: Exceeding length of sequences in alignment.");

  /* iterate over sequences */
  for (i=0; i<ali->num; i++)
    {
      /* move memory */
      memmove (ali->sequences[i]+col, ali->sequences[i]+col+1, (size_t)(ali->len-col-1 /* plus EOS */ +1)*sizeof (char));
      memmove (ali->sinfo[i]+col, ali->sinfo[i]+col+1, (size_t)(ali->len-col-1)*sizeof (SINFO));

      /* realloc memory */
      REALLOC (ali->sequences[i], char, ali->len-1 /* plus EOS */ +1);
      REALLOC (ali->sinfo[i], SINFO, ali->len-1);
    }

  ali->len--;
}

/*******************************************************
convert sequence to alignment structure
*******************************************************/
ALI *seq2Ali (char *descriptor, char *sequence, int id, BOOL usestruc)
{
  ALI *ali=NULL;

  /* allocate memory */
  MALLOC (ali, ALI, 1);

  /* copy elementary data */
  ali->id=id;
  ali->num=1;
  ali->score=0.0f;
  ali->len=strlen (sequence);

  /* alocate memory for pointers */
  MALLOC (ali->descriptors, char*, 1);
  MALLOC (ali->sequences, char*, 1);
  MALLOC (ali->sids, int, 1);
  MALLOC (ali->sinfo, SINFO*, 1);

  /* copy sequence id */
  ali->sids[0]=id;

  /* allocate memory for descriptors and sequences */
  MALLOC (ali->descriptors[0], char, strlen (descriptor)+1);
  MALLOC (ali->sequences[0], char, ali->len+1);

  /* copy descriptors, sequences and structural information */
  strncpy (ali->descriptors[0], descriptor, strlen (descriptor)+1);
  strncpy (ali->sequences[0], sequence, ali->len+1);

  /* get structural information from sequence */
  ali->sinfo[0]=getStruct (sequence, usestruc);

  return (ali);
}

/**********************************************************
convert alignment to sequence structure
**********************************************************/
SEQ ali2Seq (ALI *ali)
{
  int i, j;
  int len, offset;
  SEQ seqs;

  /* set basic structure fields */
  seqs.num=ali->num;
  seqs.type=SEQ_AUTO;

  /* allocate memory for pointers */
  MALLOC (seqs.descriptors, char*, ali->num);
  MALLOC (seqs.sequences, char*, ali->num);

  /* copy content */
  for (i=0; i<seqs.num; i++)
    {
      /* copy descriptor */
      len=strlen (ali->descriptors[i]);
      MALLOC (seqs.descriptors[i], char, len /* plus EOS */ +1);
      strncpy (seqs.descriptors[i], ali->descriptors[i], len+1);

      /* copy sequence */
      len=strlen (ali->sequences[i]);
      MALLOC (seqs.sequences[i], char, len /* plus EOS */ +1);
      strncpy (seqs.sequences[i], ali->sequences[i], len+1);
    }

  /* remove gaps */
  for (i=0; i<seqs.num; i++)
    {
      len=strlen (seqs.sequences[i]);
      offset=0;
      for (j=0; j+offset<len /* plus EOS */ +1; j++)
	{
	  while (seqs.sequences[i][j+offset]==GAPCHAR) offset++; /* search next non-gap character (string must be EOS terminated!) */
	  if (offset) seqs.sequences[i][j]=seqs.sequences[i][j+offset]; /* copy next non-gap character to current position */
	}
      REALLOC (seqs.sequences[i], char, len-offset /* plus EOS */ +1); /* resize memory allocation */
    }

  return (seqs);
}

/*********************************************************
returns an alignment projection to a subset of sequences
**********************************************************/
ALI *projectAliToSeq (ALI *ali, int *sids2, int num2)
{
  int i, i1, i2;
  int *sids1=ali->sids;
  int num1=ali->num;
  int num=0;
  int *idx;
  ALI *newali=NULL;
  size_t deslen;

  MAKE1DINT (idx, num2);

  /* find indices of sequences that match to sequence identifiers */
  for (i2=0; i2<num2; i2++) for (i1=0; i1<num1; i1++)
    if (sids1[i1]==sids2[i2]) idx[num++]=i1;
  
  if (num!=num2) FATAL ("Sequences given in partialAli not found or multiple sequence ids!");

  /* allocate memory for new alignment */
  MALLOC (newali, ALI, 1);

  /* copy elementary values */
  newali->id=ali->id;
  newali->num=num;
  newali->len=ali->len;
  newali->score=0;
  
  /* alloacte memory for pointers to descriptors, sequences, sids, structural information */
  MALLOC (newali->descriptors, char*, newali->num);
  MALLOC (newali->sequences, char*, newali->num);
  MALLOC (newali->sids, int, newali->num);
  MALLOC (newali->sinfo, SINFO*, newali->num);
  
  /* for every sequence in alignment */
  for (i=0; i<newali->num; i++)
    {
      /* descriptor */
      deslen=strlen (ali->descriptors[idx[i]]);
      MALLOC (newali->descriptors[i], char, deslen+1);
      strncpy (newali->descriptors[i], ali->descriptors[idx[i]], deslen+1);
      
      /* sequence itself */
      MALLOC (newali->sequences[i], char, newali->len+1);
      strncpy (newali->sequences[i], ali->sequences[idx[i]], newali->len+1);

      /* structural information */
      MALLOC (newali->sinfo[i], SINFO, newali->len);
      memcpy (newali->sinfo[i], ali->sinfo[idx[i]], (size_t)(newali->len)*sizeof (SINFO));

      /* sequence ids */
      newali->sids[i]=ali->sids[idx[i]];
    }

  /* free memory */
  FREE (idx);

#ifdef DEBUG
  fprintf (debugfile, "partialAli: copy sequences with these %d ids\n", num2);
  for (i=0; i<num2; i++) fprintf (debugfile, "%d ", sids2[i]);
  fprintf (debugfile, "\n");
  fprintf (debugfile, "resulting alignment is\n");
  fprintf (debugfile, "ali: #seqs=%d, len=%d\n", newali->num, newali->len);
  for (i=0; i<newali->num; i++) fprintf (debugfile, "%20s: %s\n", newali->descriptors[i], newali->sequences[i]);
#endif

  return (newali);
}

/*******************************************************
mark a special columns of an alignment with a given
character
*******************************************************/
void markAliColumns (ALI *ali, int *idx, int numidx)
{
  int i, j;
  
  for (i=0; i<numidx; i++) for (j=0; j<ali->num; j++) 
    ali->sequences[j][idx[i]]='*';
}

/********************************************************
sort alignment to order of input sequences
********************************************************/
void sortAli (ALI *ali, SEQ *seqs)
{
  int i, j;
  int num=ali->num;

  char *charptr;
  int intval;
  SINFO *sinfoptr;

  /* check integrity */
  if (num!=seqs->num) FATAL ("sortAli: Number of sequences in alignment does not match number of read sequences.");
  
  /* bubble sort sequences on sequence id's (order of sequence id's reflect order of loaded sequence) */
  for (i=1; i<=num; i++)
    {
      for (j=0; j<num-i; j++)
	{
	  if (ali->sids[j]>ali->sids[j+1])
	    {
	      SWAP (ali->descriptors[j], ali->descriptors[j+1], charptr);
	      SWAP (ali->sequences[j], ali->sequences[j+1], charptr);
	      SWAP (ali->sids[j], ali->sids[j+1], intval);
	      SWAP (ali->sinfo[j], ali->sinfo[j+1], sinfoptr);
	    }
	}
    }
}

/********************************************************
append an alignment structure to another
********************************************************/
void appendAli (ALI *dest, ALI *src)
{
  int i;

  /* check integrity */
  if (dest->num!=src->num) FATAL ("appendAli: different number of sequences.");
  for (i=0; i<dest->num; i++) if (dest->sids[i]!=src->sids[i]) FATAL ("appendAli: different order of sequences.");

  /* append sequences and structural information */
  for (i=0; i<dest->num; i++)
    {
      /* sequences */
      REALLOC (dest->sequences[i], char, dest->len+src->len+1);
      strcat (dest->sequences[i], src->sequences[i]);
      
      /* structural information */
      REALLOC (dest->sinfo[i], SINFO, dest->len+src->len);
      memcpy (dest->sinfo[i]+dest->len, src->sinfo[i], (size_t)src->len*sizeof (SINFO));
    }
  dest->len+=src->len;
}

/********************************************************
deletes an alignment structures and all sequences
contained in it
********************************************************/
void freeAli (ALI *ali)
{
  int i;

  /* get data */
  int n=ali->num;

  /* delete array of sequence id's */
  free (ali->sids);

  /* delete descriptors */
  for (i=0; i<n; ++i) free (ali->descriptors[i]);
  free (ali->descriptors);

  /* delete sequences */
  for (i=0; i<n; ++i) free (ali->sequences[i]);
  free (ali->sequences);

  /* delete structural information */
  for (i=0; i<n; i++) free (ali->sinfo[i]);
  free (ali->sinfo);

  /* free alignment structure */
  free (ali);
}

#if CHECKALI
/********************************************************
check alignment integrity
********************************************************/
void checkAli (ALI *ali)

  /* check alignment integrity */
  /* sid's */
  if (Options.Sortali) /* alignment has to be sorted to check sid's */
    for (i=0; i<ali->num-1; i++) if (ali->sids[i]!=ali->sids[i+1]-1) FATAL ("Error in sequence id's.");
  /* alignment length */
  len=strlen (ali->sequences[0]);
  if (len!=ali->len) FATAL ("Aligment length does not match specified value.");
  for (i=1; i<ali->num; i++) if (strlen (ali->sequences[i])!=len) FATAL ("Length of sequences in alignment differ.");
  /* sequences */
  seqs2=ali2Seq (ali);
  if (Seqs.num!=seqs2.num) FATAL ("Numbers differ.");
  for (i=0; i<Seqs.num; i++) 
    {
      if (Options.Verbosity>1)
	{
	  printf ("   input: %s\n", Seqs.sequences[i]);
	  printf ("from ali: %s\n", seqs2.sequences[i]);
	  printf ("     ali: %s\n\n", ali->sequences[i]);
	}
      if (strcmp (Seqs.sequences[i], seqs2.sequences[i])!=0) FATAL ("Sequences and alignment differ.");
    }
}
#endif
