#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "global.h"
#include "macros.h"
#include "options.h"
#include "ali.h"
#include "ali3.h"
#include "ali3core.h"
#include "dc.h"
#include "scores.h"
#include "split.h"
#include "purge.h"
#include "weights.h"
#include "countgaps.h"
#include "util.h"

#ifdef DEBUG
extern FILE *debugfile;
#endif

#if COUNTGAPS
extern ALI **savedalis;
extern int numsaved;
#endif

extern OPTION Options;
extern SEQ Seqs;

/*******************************************************
this is the recurrence function to create the alignment
*******************************************************/
static ALI *ali3recurrence (ALI *ali1, ALI *ali2, ALI *ali3, int s1, int l1, int s2, int l2, int s3, int l3,
			    float **Sxy, float **Sxz, float **Syz, SCORING *scoring, float *W, int depth)
{
  int i;
  int c1, c2, c3;
  int sa1, sa2, sa3, sb1, sb2, sb3;
  int la1, la2, la3, lb1, lb2, lb3;
  float **Cxy, **Cxz, **Cyz;
  static ALI *ali;
  ALI *tmpali;
  int lmax=MAX3 (l1, l2, l3);

  /* set empty alignment on top level of recursion */
  if (!depth) ali=NULL;

  if ((Options.DCLength) && (lmax>Options.DCLength))
    {
      /* split sequences */
      /*******************/
      
      /* calculate additional cost matrices */
      Cxy=calcaddcost (ali1, ali2, s1, l1, s2, l2, Sxy, scoring, W);
      Cxz=calcaddcost (ali1, ali3, s1, l1, s3, l3, Sxz, scoring, W);
      Cyz=calcaddcost (ali2, ali3, s2, l2, s3, l3, Syz, scoring, W);

      /* calculate optimal splits */
      calc3slicing (Cxy, Cxz, Cyz, l1, l2, l3, &c1, &c2, &c3);
      
      /* calculate starting positions and length of partial sequences */
      /* first part */
      sa1=s1;
      sa2=s2;
      sa3=s3;
      la1=c1;
      la2=c2;
      la3=c3;
      
      /* second part */
      sb1=sa1+la1;
      sb2=sa2+la2;
      sb3=sa3+la3;
      lb1=l1-la1;
      lb2=l2-la2;
      lb3=l3-la3;

      printf ("   -> Slicing at position (%d, %d, %d), length is (%d, %d, %d)\n", c1, c2, c3, l1, l2, l3);
      
      /* free additional scoring matrices */
      FREE2D (Cxy, l1+1, i);
      FREE2D (Cxz, l1+1, i);
      FREE2D (Cyz, l2+1, i);
      
      /* recurrence */
      /**************/
      /* prefix */
      ali3recurrence (ali1, ali2, ali3, sa1, la1, sa2, la2, sa3, la3, Sxy, Sxz, Syz, scoring, W, depth+1);
      /* suffix */
      ali3recurrence (ali1, ali2, ali3, sb1, lb1, sb2, lb2, sb3, lb3, Sxy, Sxz, Syz, scoring, W, depth+1);
    }
  else
    {
      /* longest sequence is short enough -> align sequences */
      tmpali=do3align (ali1, ali2, ali3, s1, l1, s2, l2, s3, l3, Sxy, Sxz, Syz, scoring);
      
      /* add tmpali to previous parts of alignments */
      if (!ali) ali=tmpali; /* alignment empty so far -> this is the first piece */
      else 
	{
	  appendAli (ali, tmpali); /* add tmpali */
	  freeAli (tmpali);
	}
    }
  return (ali);
}

/******************************************************
align three alignments
******************************************************/
static ALI *align3Ali (ALI *ali1, ALI *ali2, ALI *ali3, SCORING *scoring, float *W)
{
  int i;
#ifdef DEBUG
  int j, k;
#endif
  ALI *ali=NULL;
  float **Sxy, **Sxz, **Syz;

#ifdef DEBUG
  fprintf (debugfile, "align3Ali: aligning following three alignments...\n");
  fprintf (debugfile, "ali1: #seqs=%d, len=%d\n", ali1->num, ali1->len);
  for (i=0; i<ali1->num; i++) fprintf (debugfile, "%s: %s\n", ali1->descriptors[i], ali1->sequences[i]);
  for (i=0; i<ali1->num; i++)
    {
      for (j=0; j<ali1->len; j++) fprintf (debugfile, "(%6.4f %6.4f %6.4f) ", ali1->sinfo[i][j].dnpaired, ali1->sinfo[i][j].uppaired, ali1->sinfo[i][j].unpaired);
      fprintf (debugfile, "\n");
    }
  fprintf (debugfile, "ali2: #seqs=%d, len=%d\n", ali2->num, ali2->len);
  for (i=0; i<ali2->num; i++) fprintf (debugfile, "%s: %s\n", ali2->descriptors[i], ali2->sequences[i]);
  for (i=0; i<ali2->num; i++)
    {
      for (j=0; j<ali2->len; j++) fprintf (debugfile, "(%6.4f %6.4f %6.4f) ", ali2->sinfo[i][j].dnpaired, ali2->sinfo[i][j].uppaired, ali2->sinfo[i][j].unpaired);
      fprintf (debugfile, "\n");
    }
  fprintf (debugfile, "ali3: #seqs=%d, len=%d\n", ali3->num, ali3->len);
  for (i=0; i<ali3->num; i++) fprintf (debugfile, "%s: %s\n", ali3->descriptors[i], ali3->sequences[i]);
  for (i=0; i<ali3->num; i++)
    {
      for (j=0; j<ali3->len; j++) fprintf (debugfile, "(%6.4f %6.4f %6.4f) ", ali3->sinfo[i][j].dnpaired, ali3->sinfo[i][j].uppaired, ali3->sinfo[i][j].unpaired);
      fprintf (debugfile, "\n");
    }
#endif

#ifdef DEBUG
  fprintf (debugfile, "align3ali: used scoring modell...\n");
  fprintf (debugfile, "mattype: %d\n", scoring->mattype);
  fprintf (debugfile, "seqtype: %d\n", scoring->seqtype);
  fprintf (debugfile, "gaptype: %d\n", scoring->gaptype);
  fprintf (debugfile, "go: %5.2f\n", scoring->go);
  fprintf (debugfile, "ge: %5.2f\n", scoring->ge);
  fprintf (debugfile, "termgapweight: %d\n", scoring->termgapweight);
  fprintf (debugfile, "M:\n");
  fprintf (debugfile, "%6s ", "");
  for (j=0; j<NUMLET; j++) fprintf (debugfile, "%6c ", (char)j+OFFSET);
  fprintf (debugfile, "\n");
  for (i=0; i<NUMLET; i++)
    {
      fprintf (debugfile, "%6c ", (char)i+OFFSET);
      for (j=0; j<NUMLET; j++)
	{
	  fprintf (debugfile, "%6.2f ", scoring->M[i][j]);
	}
      fprintf (debugfile, "\n");
    }
#endif  
  
  /* calculate pairwise scores between all sequence position combinations */
  Sxy=calcPairwiseScore (ali1, ali2, scoring, W);
  Sxz=calcPairwiseScore (ali1, ali3, scoring, W);
  Syz=calcPairwiseScore (ali2, ali3, scoring, W);
  
#if DEBUG
  fprintf (debugfile, "align3ali: Sxy\n");
  for (i=0; i<ali1->len; ++i)
    {
      for (j=0; j<ali2->len; ++j)
	{
	  fprintf (debugfile, " %5.1f", Sxy[i][j]);
	}
      fprintf (debugfile, "\n");
    }
  fprintf (debugfile, "align3ali: Sxz\n");
  for (i=0; i<ali1->len; ++i)
    {
      for (k=0; k<ali3->len; ++k)
	{
	  fprintf (debugfile, " %5.1f", Sxz[i][k]);
	}
      fprintf (debugfile, "\n");
    }
  fprintf (debugfile, "align3ali: Syz\n");
  for (j=0; j<ali2->len; ++j)
    {
      for (k=0; k<ali3->len; ++k)
	{
	  fprintf (debugfile, " %5.1f", Syz[j][k]);
	}
      fprintf (debugfile, "\n");
    }
#endif

  /* create new alignment */
  ali=ali3recurrence (ali1, ali2, ali3, 0, ali1->len, 0, ali2->len, 0, ali3->len, Sxy, Sxz, Syz, scoring, W, 0);

#ifdef DEBUG
  fprintf (debugfile, "align3Ali: resulting alignment is...\n");
  fprintf (debugfile, "ali: #seqs=%d, len=%d\n", ali->num, ali->len);
  for (i=0; i<ali->num; i++) fprintf (debugfile, "%20s: %s\n", ali->descriptors[i], ali->sequences[i]);
#endif

  /* free memory */
  FREE2D (Sxy, ali1->len, i);
  FREE2D (Sxz, ali1->len, i);
  FREE2D (Syz, ali2->len, i);

  return (ali);
}

/*******************************************************
align the given sequences and use the phylogenetic
network as guide
*******************************************************/
ALI *alignSequences (LIST *order, float **D, SCORING *scoring)
{
  int i;
  int num1, num2, num3, num;
  int numdoubles;
  int *doubles1, *doubles2;
  int numdelcol1, numdelcol2;
  ALI **alignments;
  ALI *ali1, *ali2, *ali3;
  ALI *newali=NULL, *tmpali1, *tmpali2;
  float *W;
  int *nodes;
  int x, y, z, u, v;
  int todo, done;
  int maxnum=3*Seqs.num;
  float perc;
  int *delidx1=NULL, *delidx2=NULL;
#if COUNTGAPS
  ALI *savedali1, *savedali2;
#endif

#if DEBUG
  fprintf (debugfile, "alignSequences:\n");
  fprintf (debugfile, "mattype: %d\n", scoring->mattype);
  fprintf (debugfile, "seqtype: %d\n", scoring->seqtype);
  fprintf (debugfile, "gaptype: %d\n", scoring->gaptype);
  fprintf (debugfile, "go: %5.2f\n", scoring->go);
  fprintf (debugfile, "ge: %5.2f\n", scoring->ge);
  fprintf (debugfile, "psi: %5.2f\n", scoring->psi);
  fprintf (debugfile, "usestruc: %d\n", scoring->usestruc);
  fprintf (debugfile, "termgapweight: %d\n", scoring->termgapweight);
#endif

  /* allocate memory for alignments */
  MALLOC (alignments, ALI*, maxnum);

  /* allocate memory for weights */
  MAKE1DFLOAT (W, Seqs.num);

  /* convert the sequences to alignment structures and get structural information */
  for (i=0; i<Seqs.num; i++) alignments[i]=seq2Ali (Seqs.descriptors[i], Seqs.sequences[i], i, scoring->usestruc);

  /* pop nodes from the guide tree to determine the alignment order */
  todo=LISTnum (order);
  done=0;
  while (!LISTempty (order))
    {
      /* get the nodes/sequence ids from the guide network */
      nodes=LISTpop (order);
      x=nodes[0];
      y=nodes[1];
      z=nodes[2];
      u=nodes[3];
      v=nodes[4];
      free (nodes);

      /* set up pointer to alignments */
      ali1=alignments[x];
      ali2=alignments[y];
      ali3=alignments[z];

      /* get number of sequences of the three alignments */
      num1=ali1->num;
      num2=ali2->num;
      num3=ali3->num;
      num=num1+num2+num3;

      /* calculate sequence weights */
      memset (W, 0, (Seqs.num)*sizeof (float));
      calcWeights (W, D, ali1, ali2, ali3);
      /* align x, y and z and build a temporary alignment XYZ */
      newali=align3Ali (ali1, ali2, ali3, scoring, W);

      /* update progress counter */
      done++;
      perc=(float)done/(float)todo*100.0f;
      if (Options.Verbosity) printf ("%5.1f%% done...\r", perc);
      fflush (stdout);

      /* alignments x, y and z are not needed any further */
      freeAli (ali1);
      freeAli (ali2);
      freeAli (ali3);

      /* if the stack is empty the final alignment is in tmpali */
      /* no further steps are neccessarry */
      if (!LISTempty (order))
	{
	  /* split XYZ into XY and YZ */
	  /* splitting provides a possibility to remove introduces and possibiliy bad gaps */
	  MALLOC (tmpali1, ALI, 1);
	  MALLOC (tmpali2, ALI, 1);
	  MAKE1DINT (doubles1, num);
	  MAKE1DINT (doubles2, num);
	  splitAli (tmpali1, tmpali2, newali, u, v, num1, num2, num3, doubles1, doubles2, &numdoubles);

	  /* free entire alignment */
	  freeAli (newali);
	  
	  /* remove sequences occuring in both of the alignment splits */
	  purge (tmpali1, tmpali2, doubles1, doubles2, numdoubles, scoring, W);
			
	  /* free memory */
	  FREE1D (doubles1);
	  FREE1D (doubles2);

#if COUNTGAPS
	  /* store intermediate alignments */
	  savedali1=copyAli (tmpali1);
	  savedali2=copyAli (tmpali2);
#endif
	  
	  /* erase columns of gaps */
	  numdelcol1=eraseGapCols (tmpali1, &delidx1);
	  numdelcol2=eraseGapCols (tmpali2, &delidx2);

	  /* insert the two new alignments into the alignment array */
	  alignments[u]=tmpali1;
	  alignments[v]=tmpali2;

#if COUNTGAPS
	  /* mark columns where gaps are removed */
	  markAliColumns (savedali1, delidx1, numdelcol1);
	  markAliColumns (savedali2, delidx2, numdelcol2);
	  
	  /* save these alignments to analyze them later to get number of wrongly inserted gaps */
	  REALLOC (savedalis, ALI*, numsaved+2);
	  savedalis[numsaved++]=savedali1;
	  savedalis[numsaved++]=savedali2;
#endif
	  /* clean up */
	  free (delidx1);
	  free (delidx2);
	}
    }

  /* the alignment steps have finished here */
  /* tmpali1 (u) and tmpali2 (v) are the new */
  /* partial alignments which are constructed */
  /* from tmpali, so tmpali is the final alignment */

  /* calculate alignment score */
  calcAliScore (newali, scoring, NULL);
  
  return (newali);
}
