#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "global.h"
#include "cmdline.h"
#include "options.h"
#include "macros.h"
#include "seqio.h"
#include "distance.h"
#include "neighbornet.h"
#include "ali3.h"
#include "scores.h"
#include "scoring.h"
#include "matrices.h"
#include "countgaps.h"

#if HAVE_CONFIG_H
#include "../config.h"
#endif

#if COUNTGAPS
ALI **savedalis=NULL; /* intermediate alignments used to calculate final alignment */
int numsaved=0; /* number of intermediate alignments */
#endif

#ifdef DEBUG
FILE *debugfile=NULL;
#endif

OPTION Options; /* all needed parameters */
SEQ Seqs; /* loaded sequences */

/********************************************************
only score calculation needed
********************************************************/
static void scoreOnly (void)
{
  ALI *ali;
  SCORING *A_scoring;

  /* load the alignment from the external file */
  ali=loadAliClustal (Options.Infilename);
  
  /* allocate memory for scoring model */
  A_scoring=makeScoring ();
  
  /* setup scoring terms */
  setupScoring (NULL, A_scoring);
  
  /* determine scoring matrix */
  setScoringMatrixAli (A_scoring, 0.0f);
  
  /* calculate score of alignment */
  calcAliScore (ali, A_scoring, NULL);
  
  /* print score */
  printf ("Num: %d Len: %d, Score: %f\n", ali->num, ali->len, ali->score);
}

/********************************************************
only mean distance calculation needed
********************************************************/
static void distOnly (void)
{
  SCORING *D_scoring;
  float **PWdist;
  float meandist;

  /* read the sequences from file */
  printf ("Loading sequences from file.\n");
  loadSequences (Options.Infilename);
  
  /* allocate memory for scoring model */
  D_scoring=makeScoring ();
  
  /* setup scoring terms */
  setupScoring (D_scoring, NULL);
  
  /* determine simple scoring matrix */
  setScoringMatrixDist (D_scoring);
  
  /* compute identities and distances of the sequences */
  printf ("Computing sequence distances.\n");
  PWdist=calcSequenceDistances (Seqs.sequences, Seqs.num, D_scoring, Options.Numshuffles);
  
  /* convert pairwise distances to mean sequence distances */
  meandist=calcMeanSequenceDistance (PWdist, Seqs.num);
  
  /* print to screen */
  printf ("Mean pairwise distance of all given sequences is %6.4f.\n", meandist);
  
  /* done */
  exit (EXIT_SUCCESS);
}

/********************************************************
alignment calculation
********************************************************/
static void alignment (int argc, char *argv[])
{
  SCORING *D_scoring, *A_scoring;
  float **PWdist;
  float meandist, meanid;
  ALI *ali;
  LIST *order;

#if CHECKALI
  int len;
  SEQ seqs2;
#endif

#if COUNTGAPS
  int deletedcorrect=0;
  int deletedwrong=0;
#endif

  /* read the sequences from file */
  /* show parameter info */
  showInfo ();

  /* load sequences */
  if (Options.Verbosity) printf ("Loading sequences from file.\n");
  loadSequences (Options.Infilename);
  if (Options.Verbosity) printf ("%d sequences loaded.\n", Seqs.num);
  
  /* allocate memory for scoring model */
  D_scoring=makeScoring ();
  A_scoring=makeScoring ();

  /* setup scoring terms */
  setupScoring (D_scoring, A_scoring);

  /* save parameter settings */
  saveSettings (argc, argv);

  /* load simple scoring matrix for distance calculation (use a simple matrix for distance calculation) */
  setScoringMatrixDist (D_scoring);
 
  /* compute the identities and distances of the sequences for the NeighborNet algorithm */
  if (Options.Verbosity) printf ("Computing sequence distances.\n");
  PWdist=calcSequenceDistances (Seqs.sequences, Seqs.num, D_scoring, Options.Numshuffles);

  /* compute the alignment order */
  if (Options.Verbosity) printf ("Computing alignment order.\n");
  order=calcAliOrder (PWdist, Seqs.num);

  /* convert pairwise distances to mean sequence distances */
  meandist=calcMeanSequenceDistance (PWdist, Seqs.num);
  
  /* convert mean distance to mean identity */
  meanid=1.0f-meandist;

  /* determine the scoring matrix using calculated distances */
  setScoringMatrixAli (A_scoring, meanid);

  /* make the alignments */
  if (Options.Verbosity) printf ("Aligning sequences.\n");
  ali=alignSequences (order, PWdist, A_scoring);

#if COUNTGAPS
  /* calculate fraction of correctly deleted gaps in splitting process */
  calcwronggaps (savedalis, numsaved, ali, &deletedcorrect, &deletedwrong);
  if (Options.Verbosity)
    {
      printf ("gaps that have been deleted correctly: %d\n", deletedcorrect);
      printf ("gaps that have been deleted in-correctly: %d\n", deletedwrong);
    }
#endif

  /* sort alignment */
  if (Options.Sortali) sortAli (ali, &Seqs);

#if CHECKALI
  /* check alignment integrity */
  checkAli (ali);
#endif

  /* output alignment */
  if (Options.Verbosity) printf ("Writing alignment to file %s.\n", Options.Outfilename);
  if (Options.Outtype==OUT_CLUSTAL) saveAliClustal (Options.Outfilename, ali);
  else if (Options.Outtype==OUT_MSF) saveAliMsf (Options.Outfilename, ali);
  else FATAL ("Unknown output file format.");
}  

/***************************************************************
MAIN ROUTINE (this is the entry point of execution)
***************************************************************/
int main (int argc, char* argv[])
{
#ifdef DEBUG
  debugfile=fopen ("debugfile.txt", "w");
#endif

  /* set default options */
  initOptions ();

  /* parse command line */
  parseCmdLine (argc, argv);

  /* TODO list */
  if (Options.CalcScoreOnly) 
    scoreOnly (); /* only score calculation needed */
  else if (Options.CalcDistanceOnly) 
    distOnly (); /* only mean pairwise distance calculation */
  else 
    alignment (argc, argv);

#ifdef DEBUG
  fclose (debugfile);
#endif

  return (EXIT_SUCCESS);
}
