#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "global.h"
#include "macros.h"
#include "countgaps.h"
#include "util.h"

#if COUNTGAPS
extern SEQTYPE Seqtype;

#ifdef DEBUG
extern FILE *debugfile;
#endif

/********************************************************
get number of correctly deleted gaps during alignment 
splitting process
********************************************************/
static void getwronggaps (ALI *query, ALI *final, int *deletedcorrect, int *deletedwrong)
{
  /* make alignment between query alignment and final alignment
     use a simplified scoring scheme where only identical columns receive a positive score
     negative scores or gap-penalties are not used */
  int i, j, k;
  int posi, posj;
  int len;
  int num;
  int maxscorearg;
  int cde1, cde2;
  float S1, S2, S3;
  float **DPscore;
  char **DPdir;
  char **newseqs;
  int delwrong=0;
  int delcorrect=0;
  int gap1, gap2;
  char **sequences1;
  char **sequences2;
  int n1, n2, l1, l2, m1, m2;
  char c1, c2;

  /* project final alignment onto sequences occuring in query */
  ALI *reference=projectAliToSeq (final, query->sids, query->num);
  
  /* get data */
  sequences1=query->sequences;
  sequences2=reference->sequences;
  n1=query->num;
  n2=reference->num;
  l1=query->len;
  l2=reference->len;
  m1=l1+1;
  m2=l2+1;
  
  if (n1!=n2) FATAL ("Different number of sequences in calculation of number of wrong gaps calculation.");

  /* allocate memory */
  MAKE2DFLOAT (DPscore, m1, m2, i);
  MAKE2DCHAR (DPdir, m1, m2, i);

  /* initialize matrix */
  /* init x=0, y=0 */
  DPdir[0][0]=STOP;
  /* init y=0 */
  for (i=1; i<m1; ++i) DPdir[i][0]=DIRX;
  /* init x=0 */
  for (j=1; j<m2; ++j) DPdir[0][j]=DIRY;

  /* fill matrix */
  for (i=1; i<m1; i++)
    {
      for (j=1; j<m2; j++)
	{
	  /* update DP matrices */
	  /* check for every pair of columns in the two alignments whether they are identical */
	  num=0;
	  for (k=0; k<n1; k++) 
	    {
	      c1=sequences1[k][i-1];
	      c2=sequences2[k][j-1];
	      if ((c1=='-') || (c1=='_') || (c1=='*')) gap1=1; else gap1=0; /* any gap character */
	      if ((c2=='-') || (c2=='_') || (c2=='*')) gap2=1; else gap2=0; /* any gap character */
	      cde1=c1-OFFSET;
	      cde2=c2-OFFSET;
	      
	      if ((gap1==1) && (gap2==1)) num++; /* gap in both positions */
	      else if ((gap1==0) && (gap2==0) && (cde1==cde2)) num++; /* no gap in both positions and equal residues (even unknown) */
	    }

	  S1=DPscore[i-1][j-1]+=num;
	  S2=DPscore[i-1][j];
	  S3=DPscore[i][j-1];
	  maxscorearg=MAXARG3 (S1, S2, S3);

	  /* set traceback direction */
	  /* diagonal (align: x-1, y-1) -> aligned, no gaps */
	  if (maxscorearg==0)
	    {
	      DPscore[i][j]=S1;
	      DPdir[i][j]=DIRXY;
	    }
	  /* (align: x-1, y) -> not aligned, gap in second sequence */
	  else if (maxscorearg==1)
	    {
	      DPscore[i][j]=S2;
	      DPdir[i][j]=DIRX;
	    }
	  /* (align: x, y-1) -> not aligned, gap in first sequence */
	  else 
	    {
	      DPscore[i][j]=S3;
	      DPdir[i][j]=DIRY;
	    }
	}
    }

  /* allocate memory for new sequences */
  MALLOC (newseqs, char*, n1+n2);
  for (i=0; i<n1+n2; i++) 
    MALLOC (newseqs[i], char, m1+m2);

  /* traceback */
  len=0;
  posi=l1;
  posj=l2;

  /* build the strings in reverse order */
  while (DPdir[posi][posj]!=STOP)
    {
      if (DPdir[posi][posj]==DIRXY)
	{
	  j=0; posi--; posj--;
	  for (i=0; i<n1; i++, j++) newseqs[j][len]=sequences1[i][posi]; 
	  for (i=0; i<n2; i++, j++) newseqs[j][len]=sequences2[i][posj]; 
	}
      else if (DPdir[posi][posj]==DIRX)
	{
	  j=0; posi--;
	  for (i=0; i<n1; i++, j++) newseqs[j][len]=sequences1[i][posi]; 
	  for (i=0; i<n2; i++, j++) newseqs[j][len]='_'; /* special gap character that distinguishes here inserted gaps from previous gaps */
	}
      else if (DPdir[posi][posj]==DIRY)
	{
	  j=0; posj--;
	  for (i=0; i<n1; i++, j++) newseqs[j][len]='_'; /* special gap character that distinguishes here inserted gaps from previous gaps */
	  for (i=0; i<n2; i++, j++) newseqs[j][len]=sequences2[i][posj]; 
	}
      else FATAL ("Unknown 2D traceback direction.");
      len++;
    }

  /* terminate strings */
  for (i=0; i<n1+n2; i++)
    {
      REALLOC (newseqs[i], char, len+1);
      newseqs[i][len]=EOS;
#ifdef DEBUG
      /* reversing string only neccessarry for debugging */
      stringReverse (newseqs[i]);
#endif      
    }

  /* free memory */
  /***************/
  FREE2D (DPscore, m1, i);
  FREE2D (DPdir, m1, i);

  /* alignment has been built */
  for (i=0; i<len; i++)
    {
      if (newseqs[0][i]=='*') /* deleted gap has been found */
	{
	  for (j=n1; j<n1+n2; j++)
	    {
	      if (newseqs[j][i]==GAPCHAR) /* correct gap found in final alignment -> deletion of gap was wrong */
		delwrong++;
	      else
		delcorrect++; /* actually no gap found at this position -> correctly deleted gap */
	    }
	}
    }

#ifdef DEBUG
  fprintf (debugfile, "getwronggaps: query alignment is\n");
  fprintf (debugfile, "ali: #seqs=%d, len=%d\n", query->num, query->len);
  for (i=0; i<query->num; i++) fprintf (debugfile, "(%d) %s: %s\n", query->sids[i], query->descriptors[i], query->sequences[i]);
  fprintf (debugfile, "getwronggaps: reference alignment is\n");
  fprintf (debugfile, "ali: #seqs=%d, len=%d\n", reference->num, reference->len);
  for (i=0; i<reference->num; i++) fprintf (debugfile, "(%d) %s: %s\n", reference->sids[i], reference->descriptors[i], reference->sequences[i]);
  fprintf (debugfile, "getwronggaps: alignment of both two alignments is\n");
  for (i=0; i<n1+n2; i++) fprintf (debugfile, "%s\n", newseqs[i]);
  fprintf (debugfile, "delcorrect=%d  delwrong=%d\n", delcorrect, delwrong);
  fprintf (debugfile, "\n");
#endif

  /* delete sequences */
  for (i=0; i<n1+n2; i++) free (newseqs[i]);
  free (newseqs);

  /* return ratio of correctly deleted (=wrong) gaps */
  *deletedwrong=delwrong;
  *deletedcorrect=delcorrect;
}

/********************************************************
get fraction of correctly deleted gaps in alignment
splitting process for all intermediate alignments
********************************************************/
void calcwronggaps (ALI **savedalis, int numsaved, ALI *finalali, int *deletedcorrect, int *deletedwrong)
{
  int i;
  int dc, dw;

  for (i=0; i<numsaved; i++)
    {
      getwronggaps (savedalis[i], finalali, &dc, &dw);
      *deletedcorrect+=dc;
      *deletedwrong+=dw;
    }
}

#endif
