#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "global.h"
#include "options.h"
#include "distance.h"
#include "macros.h"

#ifdef DEBUG
extern FILE *debugfile;
#endif

static BOOL randinited;

/**********************************************************
sorting criterion for qsort
***********************************************************/
static int comp (const void *v1, const void *v2)
{
  int i1=*((int*)v1+1);
  int i2=*((int*)v2+1);
  return (i1<i2 ? -1 : (i1>i2 ? +1 : 0));
} 

/**********************************************************
shuffle a sequence
**********************************************************/
static void shuffle (char *dest, char *src)
{
  static int *shuffle;
  int i;
  int len=strlen (src);
  time_t t;

  if (!randinited)
    {
      time (&t);
      srand ((unsigned)t);
      randinited=TRUE;
    }

  /* allocate memory */
  if (!shuffle) MALLOC (shuffle, int, 2*len);

  /* generate mapping */
  for (i=0; i<len; i++)
    {
      shuffle[2*i+0]=i;
      shuffle[2*i+1]=rand ();
    }
#if DEBUG
  fprintf (debugfile, "shuffle (before sorting):\n");
  for (i=0; i<len; i++) fprintf (debugfile, "%5d ", shuffle[2*i+0]);
  fprintf (debugfile, "\n");
  for (i=0; i<len; i++) fprintf (debugfile, "%5d ", shuffle[2*i+1]);
  fprintf (debugfile, "\n");
#endif

  /* sort mapping to random numbers generated */
  qsort (shuffle, len, 2*sizeof (int), comp);

  for (i=0; i<len; i++) dest[i]=src[shuffle[2*i]];

#if DEBUG
  fprintf (debugfile, "shuffle (after sorting):\n");
  for (i=0; i<len; i++) fprintf (debugfile, "%5d ", shuffle[2*i+0]);
  fprintf (debugfile, "\n");
  for (i=0; i<len; i++) fprintf (debugfile, "%5d ", shuffle[2*i+1]);
  fprintf (debugfile, "\n");
#endif
#if DEBUG
  fprintf (debugfile, "sequences:\n");
  fprintf (debugfile, "src:  %s\n", src);
  fprintf (debugfile, "dest: %s\n", dest);
#endif
  
  dest[len]=EOS;
}

/**********************************************************
pairwise align sequences
***********************************************************/
static float alignPW (char *seq1, char *seq2, SCORING *scoring)
{	
  int i, j;
  int l1, l2, m1, m2;
  int maxarg;
  int code1, code2;
  float **DP, **Ix, **Iy;
  char **Dir;
  float smax;
  int len;
  int posi, posj;

  /* get data */
  float go=scoring->go;
  float ge=scoring->ge;
  float **M=scoring->M;
  BOOL tgw=TRUE; /* terminal gaps are weighted in distance calculation! */

  l1=(int)strlen (seq1);
  l2=(int)strlen (seq2);
  m1=l1+1;
  m2=l2+1;

  /* allocate memory */
  MAKE2DFLOAT (DP,  m1, m2, i);
  MAKE2DFLOAT (Ix,  m1, m2, i);
  MAKE2DFLOAT (Iy,  m1, m2, i);
  MAKE2DCHAR  (Dir, m1, m2, i);

  /* set boundary for x=0, y=0 */
  DP [0][0]=0;
  Ix [0][0]=-LARGE;
  Iy [0][0]=-LARGE;
  Dir[0][0]=STOP;

  /* set boundary for y=0 */
  for (i=1; i<m1; i++)
    {
      DP [i][0]=(tgw ? -go-(float)(i-1)*ge : 0.0f);
      Ix [i][0]=(tgw ? -go-(float)(i-1)*ge : 0.0f);
      Iy [i][0]=-LARGE;
      Dir[i][0]=DIRX;
    }
  /* set boundary for x=0 */
  for (j=1; j<m2; j++)
    {
      DP [0][j]=(tgw ? -go-(float)(j-1)*ge : 0.0f);
      Ix [0][j]=-LARGE;
      Iy [0][j]=(tgw ? -go-(float)(j-1)*ge : 0.0f);
      Dir[0][j]=DIRY;
    }
  /* iterate over first sequence */
  for (i=1; i<m1; i++)
    {
      /* iterate over second sequence */
      for (j=1; j<m2; j++)
	{
	  code1=(int)seq1[i-1]-OFFSET;
	  code2=(int)seq2[j-1]-OFFSET;
	  if ((code1<0) || (code1>=26) || (code2<0) || (code2>=26)) DP[i][j]=DP[i-1][j-1]; /* unknown character */
	  else DP[i][j]=DP[i-1][j-1]+M[code1][code2];
	  Ix[i][j]=MAX (DP[i-1][j]-go, Ix[i-1][j]-ge);
	  Iy[i][j]=MAX (DP[i][j-1]-go, Iy[i][j-1]-ge);
	  maxarg=MAXARG3 (DP[i][j], Ix[i][j], Iy[i][j]);
	  if (maxarg==0) 
	    {
	      /* DP [i][j]=DP[i][j]; */
	      Dir[i][j]=DIRXY;
	    }
	  else if (maxarg==1) 
	    {
	      DP [i][j]=Ix[i][j];
	      Dir[i][j]=DIRX;
	    }
	  else 
	    {
	      DP [i][j]=Iy[i][j];
	      Dir[i][j]=DIRY;
	    }
	}
    }
  
  if (!tgw)
    {
      smax=0.0f;
      /* get position of maximum score */
      for (i=0; i<m1; i++) 
	{
	  if (DP[i][l2]>smax) 
	    {
	      posi=i;
	      posj=l2;
	      smax=DP[i][l2];
	    }
	}
      for (j=0; j<m2; j++) 
	{
	  if (DP[l1][j]>smax) 
	    {
	      posi=l1;
	      posj=j;
	      smax=DP[l1][j];
	    }
	}
      /* set traceback path */
      for (i=posi+1; i<m1; i++) Dir[i][l2]=DIRX;
      for (j=posj+1; j<m2; j++) Dir[l1][j]=DIRY;
    }
  else smax=DP[l1][l2];

  /* traceback to get length of alignment */
  posi=l1;
  posj=l2;
  len=0;
  while (Dir[posi][posj]!=STOP)
    {
      if (Dir[posi][posj]==DIRXY) { posi--; posj--; }
      else if (Dir[posi][posj]==DIRX) { posi--; }
      else if (Dir[posi][posj]==DIRY) { posj--; }
      else FATAL ("Unknown traceback direction for pairwise alignment.");
      len++;
    }

  /* normalize score */
  smax/=(float)len;
	
  /* free memory */
  FREE2D (DP,  m1, i);
  FREE2D (Ix,  m1, i);
  FREE2D (Iy,  m1, i);
  FREE2D (Dir, m1, i);
  
  return (smax);
}

/**********************************************************
calculate pairwise distance between a set of
sequences
***********************************************************/
float **calcSequenceDistances (char **sequences, int num, SCORING *scoring, int shuffles)
{
  int i, j, k;
  int len1, len2;
  char *seq1r, *seq2r;
  float **S, **D;
  float **randscore;
  float dist;
  float minscore, identscore;
  float frac1, frac2;

#if DEBUG
  fprintf (debugfile, "calcSequenceDistances:\n");
  fprintf (debugfile, "mattype: %d\n", scoring->mattype);
  fprintf (debugfile, "seqtype: %d\n", scoring->seqtype);
  fprintf (debugfile, "gaptype: %d\n", scoring->gaptype);
  fprintf (debugfile, "go: %5.2f\n", scoring->go);
  fprintf (debugfile, "ge: %5.2f\n", scoring->ge);
  fprintf (debugfile, "termgapweight: %d\n", scoring->termgapweight);
  fprintf (debugfile, "M:\n");
  fprintf (debugfile, "%6s ", "");
  for (j=0; j<NUMLET; j++) fprintf (debugfile, "%6c ", (char)j+OFFSET);
  fprintf (debugfile, "\n");
  for (i=0; i<NUMLET; i++)
    {
      fprintf (debugfile, "%6c ", (char)i+OFFSET);
      for (j=0; j<NUMLET; j++)
	{
	  fprintf (debugfile, "%6.2f ", scoring->M[i][j]);
	}
      fprintf (debugfile, "\n");
    }
#endif

  /* allocate memory */
  MAKE2DFLOAT (D, num, num, i);

  /* iterate pairwise over sequences and calculate score */
  /* allocate memory */
  MAKE2DFLOAT (S, num, num, i);
      
  identscore=0.0f;
  for (i=0; i<num; i++)
    {
      for (j=0; j<=i; j++) S[i][j]=alignPW (sequences[i], sequences[j], scoring); /* include S[i][i] for all i */
      identscore+=S[i][i]; /* sum up all S[i][i] scores */
    }
  identscore/=(float)num; /* normalize summed S[i][i] scores */
      
  if (!shuffles) /* calculate sequence distances using modified STRAL formula (fast, but in-accurate) */
    {
      /* get minimum score */
      minscore=LARGE;
      for (i=0; i<num; i++) for (j=0; j<i; j++) if (S[i][j]<minscore) minscore=S[i][j];
	  
      /* convert scores to distances */
      for (i=0; i<num; i++) for (j=0; j<i; j++)
	{
	  frac1=S[i][j]-minscore+0.001f;
	  frac2=identscore-minscore;
	      
	  if ((frac1>0.0f) && (frac2>0.0f))
	    dist=-logf (frac1/frac2);
	  else
	    dist=0.0f;
	      
	  D[i][j]=(D[j][i]=MAX (0.0f, dist));
	}
    }
  else /* calculate sequence distances using formula by Feng & Doolittle (slow, but accurate) */
    {
      /* calculate average randomized alignment score */
      MAKE2DFLOAT (randscore, num, num, i);
      for (i=0; i<num; i++) for (j=0; j<i; j++)
	{
	  len1=strlen (sequences[i]);
	  len2=strlen (sequences[j]);
	  MALLOC (seq1r, char, len1+1);
	  MALLOC (seq2r, char, len2+1);
	  randscore[i][j]=0.0f;
	  for (k=0; k<shuffles; k++)
	    {
	      shuffle (seq1r, sequences[i]);
	      shuffle (seq2r, sequences[j]);
	      randscore[i][j]+=alignPW (seq1r, seq2r, scoring);
	    }
	  randscore[i][j]/=(float)shuffles;
	  FREE (seq1r);
	  FREE (seq2r);
	}
	  
      /* convert scores to distances */
      for (i=0; i<num; i++) for (j=0; j<i; j++)
	{
	  frac1=S[i][j]-randscore[i][j]+0.001;
	  frac2=identscore-randscore[i][j];
	      
	  if ((frac1>0.0f) && (frac2>0.0f))
	    dist=-logf (frac1/frac2);
	  else
	    dist=0.0f;
	      
	  D[i][j]=(D[j][i]=MAX (0.0f, dist));
	}
	  
      /* free memory for random scores */
      FREE2D (randscore, num, i);
    }

#ifdef DEBUG
  fprintf (debugfile, "pairwise sequence scores\n");
  for (i=0; i<num; i++)
    {
      for (j=0; j<=i; j++)
	{
	  fprintf (debugfile, "%7.4f ", S[i][j]);
	}
      fprintf (debugfile, "\n");
    }
  fprintf (debugfile, "\n");

  fprintf (debugfile, "pairwise sequence distances\n");
  for (i=0; i<num; i++)
    {
      for (j=0; j<num; j++)
	{
	  fprintf (debugfile, "%7.4f ", D[i][j]);
	}
      fprintf (debugfile, "\n");
    }
  fprintf (debugfile, "\n");
#endif

  /* free memory */
  FREE2D (S, num, i);
  
  return (D);
}

/**********************************************************
calculate mean pairwise sequence distance from
distance matrix
***********************************************************/
float calcMeanSequenceDistance (float **D, int num)
{
  int i, j;
  float Dmean=0.0f;

  for (i=0; i<num; i++) for (j=0; j<i; j++) Dmean+=D[i][j];
  Dmean/=(float)((num*(num-1)/2));

  return (Dmean);
}
