#ifndef GLOBAL_H
#define GLOBAL_H

/********************************************************
define statements
********************************************************/

/* support for boolean values */
#define FALSE 0
#define TRUE 1

/* system settings */
#define PATHLEN 256
#define LARGE (1e+5)
#define FLARGE (1e+5f)
#define EOL '\n'
#define EOS '\0'
#define GAPCHAR '-'

#define NUM 3 /* three alignments are aligned in each step */

/* alignment specific settings */
#define NUMLET 26 /* size of sequence alphabet */
#define BUFLEN 1000 /* length of input buffer */
#define LW 50 /* alignment output sequence width */
#define OFFSET ((int)'a')

/* traceback directions */
#define STOP 0
#define DIRX 1
#define DIRY 2
#define DIRZ 3
#define DIRXY 4
#define DIRXZ 5
#define DIRYZ 6
#define DIRXYZ 7

/* fixed alignment parameters */
#define COUNTGAPS FALSE /* count number of wrongly inserted gaps */
#define SCORESTRAL FALSE /* FALSE: old sequence-structure scoring, TRUE: STRAL like sequence-structure scoring */
#define PURGEMODE 2 /* 1: brute force, 2: greedy */
#define CHECKALI 0 /* check integrity of generated alignment */
#define NUMSHUFFLES 0 /* use fast formula for sequence distance calculation */

/* parameters for 2d alignment (distance calculation) */
#define D_MATCH 1.9f /* 1.9f */ /* match score (for identity matrix) */
#define D_MISMATCH 0.0f /* mismatch score (for identity matrix) */
#define D_NUCGAPOPEN 15.0f /* gap open penalty DNA */
#define D_NUCGAPEXTEND 6.66f /* gap extension penalty DNA */
#define D_AAGAPOPEN 10.0f /* gap open penalty PROT */
#define D_AAGAPEXTEND 0.2f /* gap extension penalty PROT */
#define D_MT MAT_DEFAULT

/* parameters for 3d alignment */
#define A_MATCH 1.9f /* 1.9f */ /* match score (for identity matrix) */
#define A_MISMATCH 0.0f /* mismatch score (for identity matrix) */
#define A_NUCGAPOPEN 15.0f /* gap open penalty DNA */
#define A_NUCGAPEXTEND 6.66f /* gap extension penalty DNA */
#define A_AAGAPOPEN 10.0f /* gap open penalty PROT */
#define A_AAGAPEXTEND 0.2f /* gap extension penalty PROT */
#define A_MT MAT_DEFAULT
#define PSI 0.5f /* balance between sequence and structure score */
#define SEQWEIGHT FALSE /* use real sequence weights */
#define USESTRUC TRUE /* use structural information */
#define DNATRANSIT 0.5f /* 0.5 */ /* cost weight for purine-purine and pyrimidine-pyrimidine pairings */
#define RNAWOBBLE 0.5f /* 0.5 */ /* cost weight for RNA wobble pairings */

/* parameters for both alignment types */
#define SAVEPARMSET FALSE /* save parameter set to file? */
#define QUASINATURALGAPS FALSE /* 0 */ /* gap approximations for three way alignments */
#define TERMGAPWEIGHT FALSE /* 0 */ /* 0: unweighted terminal gaps, 1: weighted terminal gaps */
#define SORTALI TRUE /* sort alignment to order of input sequences */
#define VERBOSITY 1 /* verbosity level */
#define OUTPUT OUT_CLUSTAL /* standard output format */
#define ST SEQ_AUTO /* sequence type */
#define GT GAP_EXTENDED /* naturual gap costs */

/********************************************************
typedefs 
********************************************************/
typedef int BOOL;
typedef void* Item;

/* sequence types */
typedef enum {
  SEQ_AUTO, /* auto detect sequence type (user did not specify explicitely) */
  SEQ_DNA, /* DNA sequence */
  SEQ_RNA, /* RNA sequence (coding or non-coding) */
  SEQ_PROTEIN /* protein sequence */
} SEQTYPE;

/* scoring matrices */
typedef enum { 
  MAT_DEFAULT, /* default sequence type (user did not specify explicitely) */
  MAT_MAMMA, /* match/mismatch matrix (using different transistions (DNA/RNA) and wobble pair scorings (RNA)) */
  MAT_IUB, /* CLUSTAL IUB matrix for nucleotide sequences */
  MAT_BLOSUM, MAT_BLOSUM30, MAT_BLOSUM40, MAT_BLOSUM45, MAT_BLOSUM62, MAT_BLOSUM80,  /* BLOSUM series */
  MAT_PAM, MAT_PAM20, MAT_PAM60, MAT_PAM120, MAT_PAM250, MAT_PAM350,  /* PAM series */
  MAT_GONNET, MAT_GONNET40, MAT_GONNET80, MAT_GONNET120, MAT_GONNET160, MAT_GONNET250, MAT_GONNET300, MAT_GONNET350 /* GONNET series */
} MATTYPE;

/* aln3nn output format */
typedef enum {
  OUT_CLUSTAL, /* output in CLUSTALW format */
  OUT_MSF /* output in MSF format */
} OUTTYPE;

/* aln3nn used gap model */
typedef enum {
  GAP_2D, /* standard 2D gap model */
  GAP_SIMPLE, /* quasi-natural gap costs (O. Gotoh) */
  GAP_EXTENDED /* full natural gap costs */
} GAPTYPE;

/* all necessary parameters to construct alignment or calculate sequence distances */
typedef struct S_SCORING
{
  SEQTYPE seqtype; /* sequence type */
  MATTYPE mattype; /* type of scoring matrix */
  GAPTYPE gaptype; /* type of gap (natural or quasi-natural) */
  BOOL termgapweight; /* weight terminal gaps */
  float go; /* gap open penalty */
  float ge; /* gap extension penalty */
  float **M; /* score entries (26x26 square matrix for pairwise scores) */
  BOOL usestruc; /* use 2nd structure information */
  float psi; /* structure sequence balance term */
} SCORING;

/* secondary structure annotation */
typedef struct S_SINFO
{
  float dnpaired; /* probability that base is paired downwards */
  float uppaired; /* probability that base is paired upwards */
  float unpaired; /* probability that base is unpaired */
} SINFO;

/* informations about loaded sequences */
typedef struct S_SEQ
{
  int num; /* number of sequences */
  SEQTYPE type; /* sequence type */
  char **descriptors; /* fasta descriptors */
  char **sequences; /* sequences itself */
} SEQ;

/* informations about calculated or loaded alignments */
typedef struct S_ALI
{
  int id; /* id of this alignments structure */
  int num; /* number of sequences in the alignment */
  int len; /* length of the alignment */
  char **descriptors; /* fasta descriptors */
  char **sequences; /* sequences in the alignment */
  SINFO **sinfo; /* secondary structure information for every sequence */
  int *sids; /* ids of all sequences in the alignment */
  float score; /* score of this alignment */
} ALI;

/* command line switches and arguments */
typedef struct S_OPTIONS
{
  char Infilename[PATHLEN]; /* input file name */
  char Outfilename[PATHLEN]; /* output file name */
  BOOL CalcDistanceOnly; /* only distance calculation? */
  BOOL CalcScoreOnly; /* only score calculation of provided alignment? */
  int DCLength; /* divide & conquer length threshold */
  BOOL Saveparmset; /* save parameter settings to external file? */
  float D_Match; /* match score for distance calculation */
  float D_Mismatch; /* mismatch score for distance calculation */
  float D_Gapopen; /* gap open penalty for distance calculation */
  float D_Gapextend; /* gap extension penalty for distance calculation */
  BOOL D_Matchgiven; /* has user provided match score for distance calculation? */
  BOOL D_Mismatchgiven; /* has user provided mismatch score for distance calculation? */
  BOOL D_Gogiven; /* has user provided gap open penalty for distance calculation? */
  BOOL D_Gegiven; /* has user provided gap extension penalty for distance calculation? */
  MATTYPE D_Mattype; /* scoring matrix type for distance calculation */
  float A_Match; /* match score for alignment calculation */
  float A_Mismatch; /* mismatch score for alignment calculation */
  float A_Gapopen; /* gap open penalty for alignment calculation */
  float A_Gapextend; /* gap extension penalty for alignment calculation */
  BOOL A_Matchgiven; /* has user provided match score for alignment calculation? */
  BOOL A_Mismatchgiven; /* has user provided mismatch score for alignment calculation? */
  BOOL A_Gogiven;  /* has user provided gap open penalty for alignment calculation? */
  BOOL A_Gegiven; /* has user provided gap extension penalty for alignment calculation? */
  MATTYPE A_Mattype; /* scoring matrix type for alignment calculation */
  float Dnatransit; /* DNA transistion score factor in [0-1] (0: mismatch score, 1: match score) for alignment calculation*/
  float Rnawobble; /* DNA transistion score factor in [0-1] (0: mismatch score, 1: match score) for alignment calculation */
  float Psi; /* sequence structure balancing term */
  BOOL Termgapweight; /* weight terminal gaps? */
  int Numshuffles; /* number of sequence shuffles for distance calculation */
  BOOL Seqweight; /* weight sequences? */
  SEQTYPE Seqtype; /* sequence type */
  GAPTYPE Gaptype; /* gap type */
  BOOL Usestruc; /* use secondary structure annotation to improve alignment of ncRNA's? */
  OUTTYPE Outtype; /* output file format */
  BOOL Sortali; /* sort alignment to order of input sequences? */
  int Verbosity; /* level of verbosity, 0:no putput except errors and warnings, 1: standard verbosity, 2: enhanced verbosity */
} OPTION;

#endif
