#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "global.h"
#include "seqio.h"
#include "options.h"
#include "matrices.h"
#include "scoringmatrices.h"
#include "macros.h"

#ifdef DEBUG
extern FILE *debugfile;
#endif

extern OPTION Options;
extern SEQ Seqs;

/********************************************************
copy hard coded protein scoring matrix to current working
matrix
********************************************************/
static void selectScoringMatrix (SCORING *scoring, MATTYPE mattype)
{
  int i, j, k;
  int ci, cj;
  int len;

  char *let;
  int *mat;
  int fac;

  /* set references */
  switch (mattype)
    {
    case MAT_IUB:
      let=IUBLET; 
      mat=IUBMAT; 
      fac=IUBFAC; 
      break;
    case MAT_BLOSUM30: 
      let=BLOSUM30LET; 
      mat=BLOSUM30MAT; 
      fac=BLOSUM30FAC; 
      break;
    case MAT_BLOSUM40: 
      let=BLOSUM40LET; 
      mat=BLOSUM40MAT; 
      fac=BLOSUM40FAC; 
      break;
    case MAT_BLOSUM45: 
      let=BLOSUM45LET; 
      mat=BLOSUM45MAT; 
      fac=BLOSUM45FAC; 
      break;
    case MAT_BLOSUM62: 
      let=BLOSUM62LET; 
      mat=BLOSUM62MAT; 
      fac=BLOSUM62FAC; 
      break;
    case MAT_BLOSUM80: 
      let=BLOSUM80LET; 
      mat=BLOSUM80MAT; 
      fac=BLOSUM80FAC; 
      break;
    case MAT_PAM20: 
      let=PAM20LET; 
      mat=PAM20MAT; 
      fac=PAM20FAC; 
      break;
    case MAT_PAM60: 
      let=PAM60LET; 
      mat=PAM60MAT; 
      fac=PAM60FAC; 
      break;
    case MAT_PAM120: 
      let=PAM120LET; 
      mat=PAM120MAT; 
      fac=PAM120FAC; 
      break;
    case MAT_PAM250: 
      let=PAM250LET; 
      mat=PAM250MAT; 
      fac=PAM250FAC; 
      break;
    case MAT_PAM350: 
      let=PAM350LET; 
      mat=PAM350MAT; 
      fac=PAM350FAC; 
      break;
    case MAT_GONNET40: 
      let=GONNET40LET; 
      mat=GONNET40MAT; 
      fac=GONNET40FAC; 
      break;
    case MAT_GONNET80: 
      let=GONNET80LET; 
      mat=GONNET80MAT; 
      fac=GONNET80FAC; 
      break;
    case MAT_GONNET120: 
      let=GONNET120LET; 
      mat=GONNET120MAT; 
      fac=GONNET120FAC; 
      break;
    case MAT_GONNET160: 
      let=GONNET160LET; 
      mat=GONNET160MAT; 
      fac=GONNET160FAC; 
      break;
    case MAT_GONNET250: 
      let=GONNET250LET; 
      mat=GONNET250MAT; 
      fac=GONNET250FAC; 
      break;
    case MAT_GONNET300: 
      let=GONNET300LET; 
      mat=GONNET300MAT; 
      fac=GONNET300FAC; 
      break;
    case MAT_GONNET350: 
      let=GONNET350LET; 
      mat=GONNET350MAT; 
      fac=GONNET350FAC; 
      break;
    default:
      FATAL ("Unknown substitution matrix.");
    }

  /* set matrix */
  scoring->mattype=mattype;
  len=strlen (let);
  for (k=0, i=0; i<len; i++)
    {
      for (j=0; j<=i; j++, k++)
	{
	  ci=(int)tolower (let[i])-OFFSET;
	  cj=(int)tolower (let[j])-OFFSET;
	  if ((ci>=0) && (ci<26) && (cj>=0) && (cj<26)) scoring->M[ci][cj]=(scoring->M[cj][ci]=(float)mat[k]/(float)fac);
	  else FATAL ("Unknown letter in matrix file.");
	}
    }
}

/********************************************************
determine substitution matrix for distance calculation
********************************************************/
void setScoringMatrixDist (SCORING *scoring)
{
  int i, j;
  MATTYPE mattype;

  /* copy requested matrix type to working matrix type */
  mattype=Options.D_Mattype;

  if ((scoring->seqtype==SEQ_DNA) || (scoring->seqtype==SEQ_RNA)) /* handle DNA/RNA sequences first */
    {
      /* set default substitution matrix */
      if (mattype==MAT_DEFAULT) mattype=MAT_MAMMA;

      /* set matrix */
      if (mattype==MAT_MAMMA)
	{
	  scoring->M['a'-OFFSET]['a'-OFFSET]=Options.D_Match;
	  scoring->M['a'-OFFSET]['c'-OFFSET]=Options.D_Mismatch;
	  scoring->M['a'-OFFSET]['g'-OFFSET]=Options.D_Mismatch;
	  scoring->M['a'-OFFSET]['t'-OFFSET]=Options.D_Mismatch;
	  scoring->M['a'-OFFSET]['u'-OFFSET]=Options.D_Mismatch;
	  
	  scoring->M['c'-OFFSET]['c'-OFFSET]=Options.D_Match;
	  scoring->M['c'-OFFSET]['g'-OFFSET]=Options.D_Mismatch;
	  scoring->M['c'-OFFSET]['t'-OFFSET]=Options.D_Mismatch;
	  scoring->M['c'-OFFSET]['u'-OFFSET]=Options.D_Mismatch;
	  
	  scoring->M['g'-OFFSET]['g'-OFFSET]=Options.D_Match;
	  scoring->M['g'-OFFSET]['t'-OFFSET]=Options.D_Mismatch;
	  scoring->M['g'-OFFSET]['u'-OFFSET]=Options.D_Mismatch;
	  
	  scoring->M['t'-OFFSET]['t'-OFFSET]=Options.D_Match;
	  scoring->M['t'-OFFSET]['u'-OFFSET]=Options.D_Match;
	  
	  scoring->M['u'-OFFSET]['u'-OFFSET]=Options.D_Match;
	  
	  /* "symmetrize" matrix */
	  for (i=0; i<26; i++) for (j=i+1; j<26; j++) scoring->M[j][i]=scoring->M[i][j];

	  scoring->mattype=MAT_MAMMA;
	}      
      else if (mattype==MAT_IUB) selectScoringMatrix (scoring, mattype);
      else FATAL ("Substitution matrix not supported for nucleotide sequences.");
    }
  else if (scoring->seqtype==SEQ_PROTEIN) /* handle amino acid sequences second */
    {
      /* set default substitution matrix for amino acid sequences */
      if (mattype==MAT_DEFAULT) mattype=MAT_BLOSUM62;
      
      /* set matrix explicitely if only a matrix famlily is given */
      if (mattype==MAT_BLOSUM) mattype=MAT_BLOSUM62;
      else if (mattype==MAT_GONNET) mattype=MAT_GONNET120;
      else if (mattype==MAT_PAM) mattype=MAT_PAM120;

      if (mattype==MAT_MAMMA)
	{
	  for (i=0; i<26; i++) for (j=0; j<=i; j++)
	    {
	      if (i==j) scoring->M[i][j]=scoring->M[j][i]=Options.D_Match;
	      else scoring->M[i][j]=scoring->M[j][i]=Options.D_Mismatch;
	    }
	  scoring->mattype=MAT_MAMMA;
	}
      else if ((mattype==MAT_BLOSUM30)  || (mattype==MAT_BLOSUM40)  || 
	       (mattype==MAT_BLOSUM45)  || (mattype==MAT_BLOSUM62)  || 
	       (mattype==MAT_BLOSUM80)  ||
	       (mattype==MAT_PAM20)     || (mattype==MAT_PAM60)     ||
	       (mattype==MAT_PAM120)    || (mattype==MAT_PAM250)    ||
	       (mattype==MAT_PAM350)    || 
	       (mattype==MAT_GONNET40)  || (mattype==MAT_GONNET80)  ||
	       (mattype==MAT_GONNET120) || (mattype==MAT_GONNET160) ||
	       (mattype==MAT_GONNET250) || (mattype==MAT_GONNET300) ||
	       (mattype==MAT_GONNET350))
 	selectScoringMatrix (scoring, mattype);
      else FATAL ("Substitution matrix not supported for amino acid sequences.");
    }
  else FATAL ("Sequence type not supported.");
  if (Options.Verbosity>1)
    {
      printf ("Using ");
      switch (scoring->mattype)
	{
	case MAT_DEFAULT: 
	  printf ("DEFAULT (not specified)");
	  break;
	case MAT_MAMMA: 
	  printf ("Match/Mismatch");
	  break;
	case MAT_IUB: 
	  printf ("IUB");
	  break;
	case MAT_BLOSUM: 
	  printf ("BLOSUM family");
	  break;
	case MAT_BLOSUM30: 
	  printf ("BLOSUM30");
	  break;
	case MAT_BLOSUM40: 
	  printf ("BLOSUM40");
	  break;
	case MAT_BLOSUM45: 
	  printf ("BLOSUM45");
	  break;
	case MAT_BLOSUM62: 
	  printf ("BLOSUM62");
	  break;
	case MAT_BLOSUM80: 
	  printf ("BLOSUM80");
	  break;
	case MAT_PAM: 
	  printf ("PAM family");
	  break;
	case MAT_PAM20: 
	  printf ("PAM20");
	  break;
	case MAT_PAM60: 
	  printf ("PAM60");
	  break;
	case MAT_PAM120: 
	  printf ("PAM120");
	  break;
	case MAT_PAM250: 
	  printf ("PAM250");
	  break;
	case MAT_PAM350: 
	  printf ("PAM350");
	  break;
	case MAT_GONNET: 
	  printf ("GONNET family");
	  break;
	case MAT_GONNET40: 
	  printf ("GONNET40");
	  break;
	case MAT_GONNET80: 
	  printf ("GONNET80");
	  break;
	case MAT_GONNET120: 
	  printf ("GONNET120");
	  break;
	case MAT_GONNET160: 
	  printf ("GONNET160");
	  break;
	case MAT_GONNET250: 
	  printf ("GONNET250");
	  break;
	case MAT_GONNET300: 
	  printf ("GONNET300");
	  break;
	case MAT_GONNET350: 
	  printf ("GONNET350");
	  break;
	}
      printf (" subsititution matrix for distance calculation.\n");
    }

#ifdef DEBUG
  fprintf (debugfile, "setDistScoringMatrix\n");
  fprintf (debugfile, "%6c ", ' ');
  for (j=0; j<26; j++) fprintf (debugfile, "%6c ", (char)(j+OFFSET));
  fprintf (debugfile, "\n");
  for (i=0; i<26; i++)
    {
      fprintf (debugfile, "%6c ", (char)(i+OFFSET));
      for (j=0; j<26; j++) fprintf (debugfile, "%6.3f ", scoring->M[i][j]);
      fprintf (debugfile, "\n");
    }
#endif
}

/********************************************************
determine substitution matrix for alignment
********************************************************/
void setScoringMatrixAli (SCORING *scoring, float identity)
{
  int i, j;
  MATTYPE mattype;

  /* copy requested matrix type to working matrix type */
  mattype=Options.A_Mattype;

  /* MAMMA matrix for DNA/RNA sequences:
     score of transistion is given by match*transit+mismatch*(1-transit) and so lies in
     the intervall [mismatch..match] for transit in [0..1]
     transistions are purine-purine substitutions as well as pyrimidine-pyrimidine substitutions
     purines are: adenine (A) and guanin (G)
     pyrimidines are: cytosine (C), thymine (T) and uracil (U)
     valid transistion are therefore: A<->G, C<->T and C<->U 
     T-U pairs are treated as matches (because in DNA sequences all U's should be T's and
     in RNA sequences all T's shoule be U's) */
  
  /* MAMMA matrix for RNA sequences:
     score of wobble pairs is given by match*wobble+mismatch*(1-wobble)
     wobble pairs are: A<->C, G<->U, but G<->T pairs are treated as wobble pairs as well
     if the sequence is considered as an RNA sequences (so all T's should be U's)
     T-U pairs are treated as matches */

  if (scoring->seqtype==SEQ_DNA) /* handle DNA sequences first */
    {
      /* set default substitution matrix */
      if (mattype==MAT_DEFAULT) mattype=MAT_MAMMA;

      /* set matrix */
      if (mattype==MAT_MAMMA)
	{
	  scoring->M['a'-OFFSET]['a'-OFFSET]=Options.A_Match;
	  scoring->M['a'-OFFSET]['c'-OFFSET]=Options.A_Mismatch;
	  scoring->M['a'-OFFSET]['g'-OFFSET]=Options.A_Match*Options.Dnatransit+Options.A_Mismatch*(1.0f-Options.Dnatransit);
	  scoring->M['a'-OFFSET]['t'-OFFSET]=Options.A_Mismatch;
	  scoring->M['a'-OFFSET]['u'-OFFSET]=Options.A_Mismatch;
	  
	  scoring->M['c'-OFFSET]['c'-OFFSET]=Options.A_Match;
	  scoring->M['c'-OFFSET]['g'-OFFSET]=Options.A_Mismatch;
	  scoring->M['c'-OFFSET]['t'-OFFSET]=Options.A_Match*Options.Dnatransit+Options.A_Mismatch*(1.0f-Options.Dnatransit);
	  scoring->M['c'-OFFSET]['u'-OFFSET]=Options.A_Match*Options.Dnatransit+Options.A_Mismatch*(1.0f-Options.Dnatransit);
	  
	  scoring->M['g'-OFFSET]['g'-OFFSET]=Options.A_Match;
	  scoring->M['g'-OFFSET]['t'-OFFSET]=Options.A_Mismatch;
	  scoring->M['g'-OFFSET]['u'-OFFSET]=Options.A_Mismatch;
	  
	  scoring->M['t'-OFFSET]['t'-OFFSET]=Options.A_Match;
	  scoring->M['t'-OFFSET]['u'-OFFSET]=Options.A_Match;
	  
	  scoring->M['u'-OFFSET]['u'-OFFSET]=Options.A_Match;
	  
	  /* "symmetrize" matrix */
	  for (i=0; i<26; i++) for (j=i+1; j<26; j++) scoring->M[j][i]=scoring->M[i][j];

	  scoring->mattype=MAT_MAMMA;
	}      
      else if (mattype==MAT_IUB) selectScoringMatrix (scoring, mattype);
      else FATAL ("Substitution matrix not supported for nucleotide sequences.");
    }
  else if (scoring->seqtype==SEQ_RNA) /* handle RNA sequences first */
    {
      /* set default subsitution matrix */
      if (mattype==MAT_DEFAULT) mattype=MAT_MAMMA;

      /* set matrix */
      if (mattype==MAT_MAMMA)
	{
	  scoring->M['a'-OFFSET]['a'-OFFSET]=Options.A_Match;
	  scoring->M['a'-OFFSET]['c'-OFFSET]=Options.A_Match*Options.Rnawobble+Options.A_Mismatch*(1.0f-Options.Rnawobble); /* wobble pair */
	  scoring->M['a'-OFFSET]['g'-OFFSET]=Options.A_Match*Options.Dnatransit+Options.A_Mismatch*(1.0f-Options.Dnatransit);
	  scoring->M['a'-OFFSET]['t'-OFFSET]=Options.A_Mismatch;
	  scoring->M['a'-OFFSET]['u'-OFFSET]=Options.A_Mismatch;
	  
	  scoring->M['c'-OFFSET]['c'-OFFSET]=Options.A_Match;
	  scoring->M['c'-OFFSET]['g'-OFFSET]=Options.A_Mismatch;
	  scoring->M['c'-OFFSET]['t'-OFFSET]=Options.A_Match*Options.Dnatransit+Options.A_Mismatch*(1.0f-Options.Dnatransit);
	  scoring->M['c'-OFFSET]['u'-OFFSET]=Options.A_Match*Options.Dnatransit+Options.A_Mismatch*(1.0f-Options.Dnatransit);
	  
	  scoring->M['g'-OFFSET]['g'-OFFSET]=Options.A_Match;
	  scoring->M['g'-OFFSET]['t'-OFFSET]=Options.A_Match*Options.Rnawobble+Options.A_Mismatch*(1.0f-Options.Rnawobble); /* wobble pair */
	  scoring->M['g'-OFFSET]['u'-OFFSET]=Options.A_Match*Options.Rnawobble+Options.A_Mismatch*(1.0f-Options.Rnawobble); /* wobble pair */
	  
	  scoring->M['t'-OFFSET]['t'-OFFSET]=Options.A_Match;
	  scoring->M['t'-OFFSET]['u'-OFFSET]=Options.A_Match;
	  
	  scoring->M['u'-OFFSET]['u'-OFFSET]=Options.A_Match;

	  /* "symmetrize" matrix */
	  for (i=0; i<26; i++) for (j=i+1; j<26; j++) scoring->M[j][i]=scoring->M[i][j];

	  scoring->mattype=MAT_MAMMA;
	}      
      else if (mattype==MAT_IUB) selectScoringMatrix (scoring, mattype);
      else FATAL ("Substitution matrix not supported for nucleotide sequences.");
    }
  else if (scoring->seqtype==SEQ_PROTEIN) /* handle amino acid sequences second */
    {
      /* set default substitution matrix for amino acid sequences */
      if (mattype==MAT_DEFAULT) mattype=MAT_BLOSUM62;

      /* chose proper fitting matrix type with respect to sequence distance */
      if (mattype==MAT_BLOSUM)
	{
	  /* choose BLOSUM matrix according to sequence distances */
	  if (identity>0.80) mattype=MAT_BLOSUM80;
	  else if (identity>0.60) mattype=MAT_BLOSUM62;
	  else if (identity>0.40) mattype=MAT_BLOSUM45;
	  else mattype=MAT_BLOSUM30;
	}
      else if (mattype==MAT_GONNET)
	{
	  /* choose GONET matrix according to sequence distances */
	  if (identity>0.65) mattype=MAT_GONNET40;
	  else if (identity>0.45) mattype=MAT_GONNET80;
	  else if (identity>0.35) mattype=MAT_GONNET120;
	  else if (identity>0.25) mattype=MAT_GONNET160;
	  else if (identity>0.15) mattype=MAT_GONNET250;
	  else mattype=MAT_GONNET350;
	}
      else if (mattype==MAT_PAM)
	{
	  /* choose PAM matrix according to sequence distances */
	  if (identity>0.80) mattype=MAT_PAM20;
	  else if (identity>0.60) mattype=MAT_PAM60;
	  else if (identity>0.40) mattype=MAT_PAM120;
	  else mattype=MAT_PAM250;
	}
     
      if (mattype==MAT_MAMMA)
	{
	  for (i=0; i<26; i++) for (j=0; j<=i; j++)
	    {
	      if (i==j) scoring->M[i][j]=scoring->M[j][i]=Options.A_Match;
	      else scoring->M[i][j]=scoring->M[j][i]=Options.A_Mismatch;
	    }
	  scoring->mattype=MAT_MAMMA;
	}
      else if ((mattype==MAT_BLOSUM30)  || (mattype==MAT_BLOSUM40)  || 
	       (mattype==MAT_BLOSUM45)  || (mattype==MAT_BLOSUM62)  || 
	       (mattype==MAT_BLOSUM80)  ||
	       (mattype==MAT_PAM20)     || (mattype==MAT_PAM60)     ||
	       (mattype==MAT_PAM120)    || (mattype==MAT_PAM250)    ||
	       (mattype==MAT_PAM350)    ||
	       (mattype==MAT_GONNET40)  || (mattype==MAT_GONNET80)  ||
	       (mattype==MAT_GONNET120) || (mattype==MAT_GONNET160) ||
	       (mattype==MAT_GONNET250) || (mattype==MAT_GONNET300) ||
	       (mattype==MAT_GONNET350))
	selectScoringMatrix (scoring, mattype);
      else FATAL ("Substitution matrix not supported for nucleotide sequences.");
    }
  else FATAL ("Sequence type not supported.");

  if (Options.Verbosity>1)
    {
      printf ("Using ");
      switch (scoring->mattype)
	{
	case MAT_DEFAULT: 
	  printf ("DEFAULT (not specified)");
	  break;
	case MAT_MAMMA: 
	  printf ("Match/Mismatch");
	  break;
	case MAT_IUB: 
	  printf ("IUB");
	  break;
	case MAT_BLOSUM: 
	  printf ("BLOSUM family");
	  break;
	case MAT_BLOSUM30: 
	  printf ("BLOSUM30");
	  break;
	case MAT_BLOSUM40: 
	  printf ("BLOSUM40");
	  break;
	case MAT_BLOSUM45: 
	  printf ("BLOSUM45");
	  break;
	case MAT_BLOSUM62: 
	  printf ("BLOSUM62");
	  break;
	case MAT_BLOSUM80: 
	  printf ("BLOSUM80");
	  break;
	case MAT_PAM: 
	  printf ("PAM family");
	  break;
	case MAT_PAM20: 
	  printf ("PAM20");
	  break;
	case MAT_PAM60: 
	  printf ("PAM60");
	  break;
	case MAT_PAM120: 
	  printf ("PAM120");
	  break;
	case MAT_PAM250: 
	  printf ("PAM250");
	  break;
	case MAT_PAM350: 
	  printf ("PAM350");
	  break;
	case MAT_GONNET: 
	  printf ("GONNET family");
	  break;
	case MAT_GONNET40: 
	  printf ("GONNET40");
	  break;
	case MAT_GONNET80: 
	  printf ("GONNET80");
	  break;
	case MAT_GONNET120: 
	  printf ("GONNET120");
	  break;
	case MAT_GONNET160: 
	  printf ("GONNET160");
	  break;
	case MAT_GONNET250: 
	  printf ("GONNET250");
	  break;
	case MAT_GONNET300: 
	  printf ("GONNET300");
	  break;
	case MAT_GONNET350: 
	  printf ("GONNET350");
	  break;
	}
      printf (" subsititution matrix for alignment.\n");
    }

#ifdef DEBUG
  fprintf (debugfile, "setAliScoringMatrix\n");
  fprintf (debugfile, "%6c ", ' ');
  for (j=0; j<26; j++) fprintf (debugfile, "%6c ", (char)(j+OFFSET));
  fprintf (debugfile, "\n");
  for (i=0; i<26; i++)
    {
      fprintf (debugfile, "%6c ", (char)(i+OFFSET));
      for (j=0; j<26; j++) fprintf (debugfile, "%6.3f ", scoring->M[i][j]);
      fprintf (debugfile, "\n");
    }
#endif
}
