#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "global.h"
#include "macros.h"
#include "neighbornet.h"
#include "list.h"

#ifdef DEBUG
extern FILE *debugfile;
#endif

struct S_NETNODE 
{
  int id;
  NETNODE *nbr; /* adjacent node */
  NETNODE *ch1; /* child 1 */
  NETNODE *ch2; /* child 2 */
  NETNODE *prev; /* previous active node */
  NETNODE *next; /* next active node */
  float Rx; /* sum of distance from this node to a active nodes */
  float Sx; /* sum of the average distance between this cluster and all other clusters */
};

/*******************************************************
create a new node 
*******************************************************/
static NETNODE *mkNetNode (int id)
{
  NETNODE* node=NULL;
  
  MALLOC (node, NETNODE, 1);
  node->id=id;
  return (node);
}

/*******************************************************
agglomerate two nodes
*******************************************************/
static void agg2way (NETNODE *x, NETNODE *y)
{
  /* make x and y neighbors */
  x->nbr=y;
  y->nbr=x;
}

/*******************************************************
agglomerate three nodes
*******************************************************/
static NETNODE* agg3way (NETNODE *x, NETNODE *y, NETNODE *z, LIST *amalgs, float **D, NETNODE *NetNodes, int *num_nodes)
{
  NETNODE *u, *v, *p;

  /* agglomerate x,y and z to give two new nodes u and v */
  /* we replace x and z by u and v and remove y from the linked list */
  /* returns a pointer to the node u */
  u=mkNetNode (*num_nodes);
  ++(*num_nodes);
  u->ch1=x;
  u->ch2=y;

  v=mkNetNode (*num_nodes);
  ++(*num_nodes);
  v->ch1=y;
  v->ch2=z;

  /* replace x by u in the linked list */
  u->next=x->next;
  u->prev=x->prev;
  if (u->next!=NULL) u->next->prev=u;
  if (u->prev!=NULL) u->prev->next=u;

  /* replace z by v in the linked list */
  v->next=z->next;
  v->prev=z->prev;
  if (v->next!=NULL) v->next->prev=v;
  if (v->prev!=NULL) v->prev->next=v;

  /* remove y from the linked list */
  if (y->next!=NULL) y->next->prev=y->prev;
  if (y->prev!=NULL) y->prev->next=y->next;

  /* add an edge between u and v and add u into the list of amalgamations */
  u->nbr=v;
  v->nbr=u;

  /* update distance matrix */
  for (p=NetNodes->next; p!=NULL; p=p->next)
    {
      D[u->id][p->id]=(D[p->id][u->id]=(2.0f/3.0f)*D[x->id][p->id]+(1.0f/3.0f)*D[y->id][p->id]);
      D[v->id][p->id]=(D[p->id][v->id]=(2.0f/3.0f)*D[z->id][p->id]+(1.0f/3.0f)*D[y->id][p->id]);
    }
  D[u->id][u->id]=(D[v->id][v->id]=0.0f);

  /* add u to amalgamations */
  LISTpush (amalgs, u);

#ifdef DEBUG
  fprintf (debugfile, "agg3way: fusing %d, %d and %d to %d and %d.\n", x->id, y->id, z->id, u->id, v->id);
#endif

  return (u);
}

/*******************************************************
agglomerate four nodes
*******************************************************/
static void agg4way (NETNODE *x2, NETNODE *x, NETNODE *y, NETNODE *y2, LIST *amalgs, float **D, NETNODE *NetNodes, int *num_nodes)
{
  NETNODE *u;

  /* replace x2, x, y, y2 by two vertices performed using two 3 way amalgamations */
  /* replace x2, x, y by two nodes, equal to x2->prev->next and y->prev->next */
  u=agg3way (x2, x, y, amalgs, D, NetNodes, num_nodes);
  agg3way (u, u->nbr, y2, amalgs, D, NetNodes, num_nodes);
}

/*******************************************************
compute Rx
*******************************************************/
static float compute_Rx (NETNODE *z, NETNODE *Cx, NETNODE *Cy, float **D, NETNODE *NetNodes)
{
  float Rx=0.0f;
  NETNODE *p;

  for (p=NetNodes->next; p!=NULL; p=p->next)
    {
      if ((p==Cx) || (p==Cx->nbr) || (p==Cy) || (p==Cy->nbr) || (p->nbr==NULL))
	Rx+=D[z->id][p->id];
      else 
	Rx+=(D[z->id][p->id]/2.0f); /* we take the average of the distances */
    }
  return (Rx);
}

/*******************************************************
agglomerates nodes
*******************************************************/
static void agglom_nodes (LIST *amalgs, float **D, NETNODE *NetNodes, int *num_nodes)
{
  NETNODE *p, *q, *Cx, *Cy, *x, *y;
  float Qpq, Dpq, best=0.0f;
  int num_active, num_clusters;
  int m;

  num_active=(num_clusters=(*num_nodes));

  /* main agglomeration loop */
  while (num_active>3)
    {
      /* Special case! Avoid a divide by zero when calculating Qpg. */
      if ((num_active==4) && (num_clusters==2))
	{
	  p=NetNodes->next;
	  if (p->next!=p->nbr) q=p->next;
	  else q=p->next->next;
	  if (D[p->id][q->id]+D[p->nbr->id][q->nbr->id]<D[p->id][q->nbr->id]+D[p->nbr->id][q->id])
	    agg3way (p, q, q->nbr, amalgs, D, NetNodes, num_nodes);
	  else 
	    agg3way (p, q->nbr, q, amalgs, D, NetNodes, num_nodes);
	  break; /* leave the loop */
	}

      /* compute the averaged sums s_i from each cluster to every other cluster */
      /* two-fold speed-up due to symmetrie */
      for (p=NetNodes->next; p!=NULL; p=p->next) p->Sx=0.0f;
      for (p=NetNodes->next; p!=NULL; p=p->next)
	{
	  if ((p->nbr==NULL) || (p->nbr->id>p->id))
	    {
	      for (q=p->next; q!=NULL; q=q->next)
		{
		  if ((q->nbr==NULL) || ((q->nbr->id>q->id) && (q->nbr!=p)))
		    {
		      Dpq=0.0f;
		      if ((p->nbr==NULL) && (q->nbr==NULL)) Dpq=D[p->id][q->id];
		      else if ((p->nbr!=NULL) && (q->nbr==NULL)) Dpq=(D[p->id][q->id]+D[p->nbr->id][q->id])/2.0f;
		      else if ((p->nbr==NULL) && (q->nbr!=NULL)) Dpq=(D[p->id][q->id]+D[p->id][q->nbr->id])/2.0f;
		      else Dpq=(D[p->id][q->id]+D[p->id][q->nbr->id]+D[p->nbr->id][q->id]+D[p->nbr->id][q->nbr->id])/4.0f;
		      p->Sx+=Dpq;
		      if (p->nbr!=NULL) p->nbr->Sx+=Dpq;
		      q->Sx+=Dpq;
		      if (q->nbr!=NULL) q->nbr->Sx+=Dpq;
		    }
		}
	    }
	}

      Cx=(Cy=NULL);
      /* minimize (m-2)*D[C_i,C_k]-Sx-Sy -> get the two clusters */
      /* find the two cluster minimizing the Q criterion */
      best=FLARGE;
      for (p=NetNodes->next; p!=NULL; p=p->next)
	{
	  if ((p->nbr!=NULL) && (p->nbr->id<p->id)) continue; /* we only evaluate one node per cluster */
	  for (q=NetNodes->next; q!=p; q=q->next)
	    {
	      if ((q->nbr!=NULL) && (q->nbr->id<q->id)) continue; /* we only evaluate one node per cluster */
	      if (q->nbr==p) continue; /* we only evaluate nodes in different clusters */

	      if ((p->nbr==NULL) && (q->nbr==NULL)) Dpq=D[p->id][q->id];
	      else if ((p->nbr!=NULL) && (q->nbr==NULL)) Dpq=(D[p->id][q->id]+D[p->nbr->id][q->id])/2.0f;
	      else if ((p->nbr==NULL) && (q->nbr!=NULL)) Dpq=(D[p->id][q->id]+D[p->id][q->nbr->id])/2.0f;
	      else Dpq=(D[p->id][q->id]+D[p->id][q->nbr->id]+D[p->nbr->id][q->id]+D[p->nbr->id][q->nbr->id])/4.0f;
	      Qpq=((float)num_clusters-2.0f)*Dpq-p->Sx-q->Sx;
	      /* check, if this is the best so far */
	      if ((Cx==NULL || (Qpq<best)) && (p->nbr!=q))
		{
		  Cx=p;
		  Cy=q;
		  best=Qpq;
		}
	    }
	}

      /* find the node in each cluster */
      x=Cx;
      y=Cy;

      if ((Cx->nbr!=NULL) || (Cy->nbr!=NULL))
	{
	  Cx->Rx=compute_Rx (Cx, Cx, Cy, D, NetNodes);
	  if (Cx->nbr!=NULL) Cx->nbr->Rx=compute_Rx (Cx->nbr, Cx, Cy, D, NetNodes);
	  Cy->Rx=compute_Rx (Cy, Cx, Cy, D, NetNodes);
	  if (Cy->nbr!=NULL) Cy->nbr->Rx=compute_Rx (Cy->nbr, Cx, Cy, D, NetNodes);
	}

      m=num_clusters;
      if (Cx->nbr!=NULL) m++;
      if (Cy->nbr!=NULL) m++;

      best=((float)m-2.0f)*D[Cx->id][Cy->id]-Cx->Rx-Cy->Rx;
      if (Cx->nbr!=NULL)
	{
	  Qpq=((float)m-2.0f)*D[Cx->nbr->id][Cy->id]-Cx->nbr->Rx-Cy->Rx;
	  if (Qpq<best)
	    {
	      x=Cx->nbr;
	      y=Cy;
	      best=Qpq;
	    }
	}
      if (Cy->nbr!=NULL)
	{
	  Qpq=((float)m-2.0f)*D[Cx->id][Cy->nbr->id]-Cx->Rx-Cy->nbr->Rx;
	  if (Qpq<best)
	    {
	      x=Cx;
	      y=Cy->nbr;
	      best=Qpq;
	    }
	}
      if ((Cx->nbr!=NULL) && (Cy->nbr!=NULL))
	{
	  Qpq=((float)m-2.0f)*D[Cx->nbr->id][Cy->nbr->id]-Cx->nbr->Rx-Cy->nbr->Rx;
	  if (Qpq<best)
	    {
	      x=Cx->nbr;
	      y=Cy->nbr;
	      best=Qpq;
	    }
	}

      /* we perform an agglomeration: one of three types */
      /* both vertices are isolated -> add an edge */
      if ((x->nbr==NULL) && (y->nbr==NULL))
	{
	  agg2way (x, y);
	  num_clusters--;
	}
      /* only x is isolated */
      else if (x->nbr==NULL)
	{
	  agg3way (x, y, y->nbr, amalgs, D, NetNodes, num_nodes);
	  num_clusters--;
	  num_active--;
	}
      /* only y is isolated OR only four nodes left which are all not isolated */
      else if ((y->nbr==NULL) || (num_active==4))
	{
	  agg3way (y, x, x->nbr, amalgs, D, NetNodes, num_nodes);
	  num_clusters--;
	  num_active--;
	}
      /* both nodes are connected to others and there are more than 4 active nodes */
      else
	{
	  agg4way (x->nbr, x, y, y->nbr, amalgs, D, NetNodes, num_nodes);
	  num_clusters--;
	  num_active-=2;
	}
    }
}

/*******************************************************
expand the nodes
*******************************************************/
static void quick_expand (LIST *amalgs, NETNODE* NetNodes, LIST *guide)
{
  int *guidearray;

  NETNODE *x, *y, *z, *u, *v, *a, *tmp;

  /* set up the circular order for the first three nodes */
  x=NetNodes->next;
  y=x->next;
  z=y->next;
  z->next=x;
  x->prev=z;

  /* add nodes to guide stack */
  MAKE1DINT (guidearray, 5);
  guidearray[0]=x->id;
  guidearray[1]=y->id;
  guidearray[2]=z->id;
  guidearray[3]=0;
  guidearray[4]=0;
  LISTpush (guide, guidearray);

#ifdef DEBUG
  fprintf (debugfile, "quick_expand: add (%d,%d,%d)->(%d,%d) to guide.\n", x->id, y->id, z->id, 0, 0);
#endif

  /* now do the rest of the expansions */
  while (!LISTempty (amalgs))
    {
      /* find the three elements replacing u and v */
      u=LISTpop (amalgs);
      v=u->nbr;
      x=u->ch1;
      y=u->ch2;
      z=v->ch2;

      /* swap u and v around if v comes before u in the circular ordering being built up */
      if (v!=u->next)
	{
	  SWAP (u, v, tmp);
	  SWAP (x, z, tmp);
	}

      /* insert x, y, z into the circular order */
      x->prev=u->prev;
      x->prev->next=x;
      x->next=y;
      y->prev=x;
      y->next=z;
      z->prev=y;
      z->next=v->next;
      z->next->prev=z;

      /* add nodes to guide stack */
      MAKE1DINT (guidearray, 5);
      guidearray[0]=x->id;
      guidearray[1]=y->id;
      guidearray[2]=z->id;
      guidearray[3]=u->id;
      guidearray[4]=v->id;
      LISTpush (guide, guidearray);

#ifdef DEBUG
      fprintf (debugfile, "quick_expand: add (%d,%d,%d)->(%d,%d) to guide.\n", x->id, y->id, z->id, u->id, v->id);
#endif

      /* free unused nodes */
      free (u);
      free (v);
    }

  /* clean up */
  /* when we exit, we know that the point x points to a node in the circular order */
  /* we loop through until we find the node after the taxa zero */
  while (x->id!=0) x=x->next;

  /* free up the linked list */
  a=x->next;
  while (a!=x)
    {
      a=a->next;
      free (a->prev);
    }
  free (x);
}

/*******************************************************
NeighborNet main routine
********************************************************/
static int neighbor_net (float **inputD, int num_taxa, LIST *guide)
{
  int i, j;
  int max_num_nodes, num_nodes;
  float  **D;
  NETNODE *NetNodes, *TaxNode;
  LIST *amalgs;

  num_nodes=num_taxa; /* initially start with one node for each taxa */
  max_num_nodes=num_taxa*3-6;

  /* copy distance matrix into a larger scratch distance matrix */
  MAKE2DFLOAT (D, max_num_nodes, max_num_nodes, i);
  for (i=0; i<num_taxa; i++) for (j=0; j<num_taxa; j++) D[i][j]=inputD[i][j];

  /* allocate memory for nodes */
  NetNodes=mkNetNode (0);

  /* nodes are stored in a doubly linked list */
  /* initially all singleton nodes are active */
  for (i=num_taxa-1; i>=0; --i)
    {
      TaxNode=mkNetNode (i);
      TaxNode->next=NetNodes->next;
      NetNodes->next=TaxNode;
    }

  /* set up links in the other direction */
  for (TaxNode=NetNodes; TaxNode->next!=NULL; TaxNode=TaxNode->next)
    TaxNode->next->prev=TaxNode;

  /* create the list for node amalgamations */
  amalgs=LISTinit ();

  /* perform the agglomeration step */
  agglom_nodes (amalgs, D, NetNodes, &num_nodes);

  /* expand */
  quick_expand (amalgs, NetNodes, guide);

  /* free memory */
  FREE2D (D, max_num_nodes, i);
  LISTfree (amalgs);

  return (num_nodes);
}

/*******************************************************
compute the alignment order of triplewise progressive 
alignment (NeighborNet-Algorithm)
********************************************************/
LIST *calcAliOrder (float **D, int numseqs)
{
  LIST* guide;

  /* initialize the guide stack */
  guide=LISTinit ();

  /* compute the circular ordering */
  neighbor_net (D, numseqs, guide);

  return (guide);
}
