#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "global.h"
#include "tree.h"
#include "macros.h"

/*******************************************************
neighbor joining algorithm
returns the "root" of the NJ-tree
********************************************************/
NODE *njoin (float **distances, int *sids, int num)
{
  int i, j;
  int mini=0, minj=1;
  int newid;
  int last1id, last2id;

  float V1, V2;
  float minMij;
  float dist;

  float **D, **M;
  NODE *supnode=NULL;
  NODE *last1, *last2;
  NODE **active;
  EDGE *edge;

  /* get data */
  int maxnum=2*num;
  int activenum=num;
  int curnum=num;

  /* if only one taxon */
  if (num==1) supnode=NODEmake (sids[0]);
  /* more than one taxon */
  else
    {
      /* create array for matrices */
      MAKE2DFLOAT (D, maxnum, maxnum, i); /* local distance working matrix */
      MAKE2DFLOAT (M, maxnum, maxnum, i); /* matrix of rate corrected distances */
      
      /* create array for active nodes */
      MALLOC (active, NODE*, maxnum);
      
      /* create central supernode */
      supnode=NODEmake (-1);
      
      /* create a node for every sequence and connect it to the central supernode */
      for (i=0; i<activenum; i++)
	{
	  /* create new node */
	  active[i]=NODEmake (sids[i]);
	  /* create new edge */
	  edge=EDGEmake ();
	  /* insert edge */
	  NODEaddedge (supnode, active[i], edge);
	}
      
      /* copy distances from global matrix to local working matrix */
      for (i=0; i<num; i++)
	{
	  for (j=0; j<num; j++)
	    {
	      D[i][j]=distances[sids[i]][sids[j]];
	    }
	}
      
      /* start NJoin process */
      while (activenum>2)
	{
	  newid=curnum;
	  
	  /* calculate for every active node distance to other active nodes */
	  for (i=0; i<maxnum; i++)
	    {
	      if (active[i])
		{
		  active[i]->Rx=0.0f;
		  for (j=0; j<maxnum; j++) if ((active[j]) && (i!=j)) (active[i]->Rx)+=D[i][j];
		  active[i]->Rx/=(float)(activenum-2);
		}
	    }
	  
	  /* calculate rate corrected distance matrix and find minimum distance pairs */
	  minMij=LARGE;
	  for (i=0; i<curnum; i++) for (j=i+1; j<curnum; j++)
	    {
	      if ((active[i]) && (active[j]))
		{
		  M[i][j]=D[i][j]-((active[i]->Rx)+(active[j]->Rx));
		  if (M[i][j]<minMij)
		    {
		      minMij=M[i][j];
		      mini=i;
		      minj=j;
		    }
		}
	    }
	  
	  /* found a pair (i,j) with minimum corrected distances */
	  /* calculate branch lengths now */
	  V1=D[mini][minj]/2.0f+((active[mini]->Rx)-(active[minj]->Rx))/2.0f;
	  V2=D[mini][minj]/2.0f+((active[minj]->Rx)-(active[mini]->Rx))/2.0f;
	  
	  /* insert a new node to active nodes */
	  active[newid]=NODEmake (-1);
	  
	  /* set distance from all nodes to new node */
	  for (i=0; i<curnum; i++) D[i][newid]=(D[newid][i]=(D[mini][i]+D[minj][i]-D[mini][minj])/2.0f);
	  
	  /* increase total number of nodes */
	  curnum++;
	  
	  /* add an edge from central node to new node */
	  edge=EDGEmake ();
	  NODEaddedge (supnode, active[newid], edge);
	  
	  /* delete edges from central node to node 1 and node 2 */
	  NODEdeledge (supnode, active[mini]);
	  NODEdeledge (supnode, active[minj]);
	  
	  /* add an outgoing edge with length l from the new node to node 1 and node 2 */
	  edge=EDGEmake ();
	  edge->length=V1;
	  NODEaddedge (active[newid], active[mini], edge);
	  edge=EDGEmake ();
	  edge->length=V2;
	  NODEaddedge (active[newid], active[minj], edge);
	  
	  /* delete node i and j from active nodes */
	  active[mini]=NULL;
	  active[minj]=NULL;
	  
	  /* update number of active nodes */
	  activenum--;
	}
      
      /* now, there are only two active original nodes left */
      if (supnode->outnum!=2) FATAL ("Njoin: Number of outgoing nodes does not match.");
      
      /* get two last nodes */
      last1=supnode->out[0]->to;
      last2=supnode->out[1]->to;
      
      /* get the array indices of the two last nodes */
      for (last1id=0; last1id<curnum; last1id++) if (last1==active[last1id]) break;
      for (last2id=last1id+1; last2id<curnum; last2id++) if (last2==active[last2id]) break;
      
      /* get distance between two last nodes */
      dist=D[last1id][last2id];
      
      /* set the proper edge length between the supernode and the two nodes left */
      NODEsetedgelength (supnode, last1, dist/2.0f);
      NODEsetedgelength (supnode, last2, dist/2.0f);
      
      /* clean up */
      FREE2D (D, maxnum, i);
      FREE2D (M, maxnum, i);
      FREE (active);
    }

  return (supnode);
}
