#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "global.h"
#include "macros.h"
#include "purge.h"
#include "ali.h"
#include "scores.h"

#ifdef DEBUG
extern FILE *debugfile;
#endif

/*********************************************************
erases rows of an alignment consisting of gaps only
returns vector to positions of deleted columns
**********************************************************/
int eraseGapCols (ALI *ali, int **deleted)
{
  int i, j;
  int numdeleted=0;

  /* nothing deleted so far */
  *deleted=NULL;

  /* search for columns to be deleted, but DO NOT delete here already */
  for (i=0; i<ali->len; i++)
    {
      /* iterate over sequences at loci i */
      for (j=0; j<ali->num; j++) if (ali->sequences[j][i]!=GAPCHAR) break; /* no gap-character found, abort seach in current column */

      /* gap characters only found? */
      if (j==ali->num)
	{
	  /* store position of column to be deleted */
	  REALLOC (*deleted, int, numdeleted+1);
	  (*deleted)[numdeleted++]=i;
	}
    }

  /* delete columns (delete in reverse order to preserve indices,
     delete is a strictly monoton inceasing function) */
  for (i=numdeleted-1; i>=0; i--) deleteAliCol (ali, (*deleted)[i]);

#ifdef DEBUG
  fprintf (debugfile, "purge: alignment after erasing %d gap columns:\n", numdeleted);
  fprintf (debugfile, "ali: #seqs=%d, len=%d\n", ali->num, ali->len);
  for (i=0; i<ali->num; i++) fprintf (debugfile, "%s: %s\n", ali->descriptors[i], ali->sequences[i]);
#endif

  return (numdeleted);
}

#if PURGEMODE==1
/*********************************************************
removing double occuring sequences from the alignment
with the worst sequence-alignment score
ndid: number of duplicated sequences
dids1: position (not id!) of duplicated sequences in first 
alignment
dids2: position (not id!) of duplicated sequences in 
second alignment
NOTICE:
this implementation is the brute force method (all
combinations are tried)
a branch-and-bound implementation should be considered
**********************************************************/
void purge (ALI *ali1, ALI *ali2, int *dids1, int *dids2, int ndid, SCORING *scoring, float *W)
{
  int i, j;
  unsigned long perm, maxperm, bestperm=0;
  char bit1, bit2;
  char *umask1, *umask2;
  float score1, score2;
  float *score;
  float bestscore=0.0f;

  /* get data */
  int num1=ali1->num;
  int num2=ali2->num;

#ifdef DEBUG
  fprintf (debugfile, "purge1\n");
  fprintf (debugfile, "%d sequences are duplicated\n", ndid);
  fprintf (debugfile, "duplicated sequences in first alignment are: ");
  for (i=0; i<ndid; i++) fprintf (debugfile, "#id:%d (pos:%d) ", ali1->sids[dids1[i]], dids1[i]);
  fprintf (debugfile, "\n");
  fprintf (debugfile, "duplicated sequences in second alignment are: ");
  for (i=0; i<ndid; i++) fprintf (debugfile, "#id:%d (pos:%d) ", ali2->sids[dids2[i]], dids2[i]);
  fprintf (debugfile, "\n");
#endif

  if (ndid>30) FATAL ("Removing more then 30 duplicated sequences is impossible.");
  if ((!ndid) || (!dids1) || (!dids2)) return;

  /* maximum number of permutations to try */
  maxperm=(unsigned long)pow (2ul, (unsigned long)ndid);

  /* allocate memory */
  MAKE1DCHAR (umask1, num1);
  MAKE1DCHAR (umask2, num2);
  MAKE1DFLOAT (score, maxperm);

  /* iterate over all cominations of removing duplicated sequences from the two alignments */
  for (perm=0; perm<maxperm; perm++)
    {
      /* important! masks only these sequence out that are duplicated */
      memset (umask1, 0, (size_t)num1*sizeof (char));
      memset (umask2, 0, (size_t)num2*sizeof (char));
      for (j=0; j<ndid; j++)
	{
	  bit1=(perm>>j)&1;
	  bit2=(~bit1)&1;
	  umask1[dids1[j]]=bit1;
	  umask2[dids2[j]]=bit2;
	}
		
      score1=calcMaskAliScore (ali1, umask1, scoring, W);
      score2=calcMaskAliScore (ali2, umask2, scoring, W);
      score[perm]=score1+score2;

#ifdef DEBUG
      fprintf (debugfile, "purge1\n");
      fprintf (debugfile, "trying to remove sequences from first alignment:  ");
      for (j=0; j<ndid; j++) if (umask1[dids1[j]]) fprintf (debugfile, "%d ", ali1->sids[dids1[j]]);
      fprintf (debugfile, " - score is %.3f\n", score1);
      fprintf (debugfile, "trying to remove sequences from second alignment: ");
      for (j=0; j<ndid; j++) if (umask2[dids2[j]]) fprintf (debugfile, "%d ", ali2->sids[dids2[j]]);
      fprintf (debugfile, " - score is %.3f\n", score2);
      fprintf (debugfile, "overall score is %.3f\n", score[perm]);
#endif
    }

  /* find combination with maximum score */
  for (perm=0; perm<maxperm; perm++)
    {
      if ((!perm) || (score[perm]>bestscore))
	{
	  bestscore=score[perm];
	  bestperm=perm;
	}
    }
	
  /* re-create umasks */
  /* important! masks only these sequence out that are duplicated */
  memset (umask1, 0, (size_t)num1*sizeof (char));
  memset (umask2, 0, (size_t)num2*sizeof (char));
  for (j=0; j<ndid; j++)
    {
      bit1=(bestperm>>j)&1;
      bit2=(~bit1)&1;
      umask1[dids1[j]]=bit1;
      umask2[dids2[j]]=bit2;
    }

#ifdef DEBUG
  fprintf (debugfile, "purge1\n");
  fprintf (debugfile, "removing sequences from first alignment: ");
  for (j=0; j<ndid; j++) if (umask1[dids1[j]]) fprintf (debugfile, "%d ", ali1->sids[dids1[j]]);
  fprintf (debugfile, "\n");
  fprintf (debugfile, "removing sequences from second alignment: ");
  for (j=0; j<ndid; j++) if (umask2[dids2[j]]) fprintf (debugfile, "%d ", ali2->sids[dids2[j]]);
  fprintf (debugfile, "\n");
  fprintf (debugfile, "overall score is %.3f\n", score[bestperm]);
#endif

  /* remove disabled sequences */
  /* first alignment */
  for (i=0; i<num1;)
    {
      if (umask1[i]) /* delete sequence i */
	{
#ifdef DEBUG
	  fprintf (debugfile, "purge1: deleting sequence %d in first alignment.\n", ali1->sids[i]);
#endif
	  FREE1D (ali1->descriptors[i]);
	  FREE1D (ali1->sequences[i]);
	  FREE1D (ali1->sinfo[i]);
	  for (j=i; j<num1-1; j++) /* move all j+1 to j for j=i..n-1 */
	    {
	      ali1->sids[j]=ali1->sids[j+1];
	      ali1->descriptors[j]=ali1->descriptors[j+1];
	      ali1->sequences[j]=ali1->sequences[j+1];
	      ali1->sinfo[j]=ali1->sinfo[j+1];
	      umask1[j]=umask1[j+1];
	    }
	  num1--;
	}
      else i++;
    }
  REALLOC (ali1->sids, int, num1);
  REALLOC (ali1->descriptors, char*, num1);
  REALLOC (ali1->sequences, char*, num1);
  REALLOC (ali1->sinfo, SINFO*, num1);
  ali1->num=num1;
	
  /* second alignment */
  for (i=0; i<num2;)
    {
      if (umask2[i])
	{
#ifdef DEBUG
	  fprintf (debugfile, "purge1: deleting sequence %d in second alignment.\n", ali2->sids[i]);
#endif
	  FREE1D (ali2->descriptors[i]);
	  FREE1D (ali2->sequences[i]);
	  FREE1D (ali2->sinfo[i]);
	  for (j=i; j<num2-1; j++)
	    {
	      ali2->sids[j]=ali2->sids[j+1];
	      ali2->descriptors[j]=ali2->descriptors[j+1];
	      ali2->sequences[j]=ali2->sequences[j+1];
	      ali2->sinfo[j]=ali2->sinfo[j+1];
	      umask2[j]=umask2[j+1];
	    }
	  num2--;
	}
      else i++;
    }
  REALLOC (ali2->sids, int, num2);
  REALLOC (ali2->descriptors, char*, num2);
  REALLOC (ali2->sequences, char*, num2);
  REALLOC (ali2->sinfo, SINFO*, num2);
  ali2->num=num2;

  /* free memory */
  FREE1D (umask1);
  FREE1D (umask2);
  FREE1D (score);

#ifdef DEBUG
  fprintf (debugfile, "purge1: alignments after erasing duplicated sequences:\n");
  fprintf (debugfile, "ali1: #seqs=%d, len=%d\n", ali1->num, ali1->len);
  for (i=0; i<ali1->num; i++) fprintf (debugfile, "%s: %s\n", ali1->descriptors[i], ali1->sequences[i]);
  fprintf (debugfile, "ali2: #seqs=%d, len=%d\n", ali2->num, ali2->len);
  for (i=0; i<ali2->num; i++) fprintf (debugfile, "%s: %s\n", ali2->descriptors[i], ali2->sequences[i]);
#endif
}
#endif

#if PURGEMODE==2
/*********************************************************
removing double occuring sequences from the alignment
with the worst sequence-alignment score
ndid: number of duplicated sequences
dids1: position (not id!) of duplicated sequences in first 
alignment
dids2: position (not id!) of duplicated sequences in 
second alignment
NOTICE:
this is the greedy-method (remove the sequence with the
smallest score in the corresponding alignment first)
**********************************************************/
void purge (ALI *ali1, ALI *ali2, int *dids1, int *dids2, int ndid, SCORING *scoring, float *W)
{
  int i, j;
  char *umask1, *umask2;
  char *fromwhich;
  float *sc1, *sc2, *scM;
  int processed;
  int found;
  float selM;
  int selj;
  int numdis1, numdis2;

  /* get data */
  int num1=ali1->num;
  int num2=ali2->num;

#ifdef DEBUG
  fprintf (debugfile, "purge2\n");
  fprintf (debugfile, "%d sequences are duplicated\n", ndid);
  fprintf (debugfile, "Duplicated sequences in first alignment are: ");
  for (i=0; i<ndid; i++) fprintf (debugfile, "#id:%d (pos:%d) ", ali1->sids[dids1[i]], dids1[i]);
  fprintf (debugfile, "\n");
  fprintf (debugfile, "Duplicated sequences in second alignment are: ");
  for (i=0; i<ndid; i++) fprintf (debugfile, "#id:%d (pos:%d) ", ali2->sids[dids2[i]], dids2[i]);
  fprintf (debugfile, "\n");
#endif

  if ((!ndid) || (!dids1) || (!dids2)) return;

  /* allocate memory for marking of disabled sequences */
  MAKE1DCHAR (umask1, num1); /* disabled sequences in alignment 1 */
  MAKE1DCHAR (umask2, num2); /* disabled sequences in alignment 2 */
  MAKE1DCHAR (fromwhich, ndid);  /* in which alignment is score lower */
  MAKE1DFLOAT (sc1, ndid);
  MAKE1DFLOAT (sc2, ndid);
  MAKE1DFLOAT (scM, ndid);

  numdis1=0;
  numdis2=0;
  /* iterate and disable one sequence per iteration */
  for (i=0; i<ndid; i++)
    {
      /* updates scores of all duplicated and so far not processed sequences to active (not disabled) remainder of alignment */
      for (j=0; j<ndid; j++)
	{
	  processed=((umask1[dids1[j]]) || (umask2[dids2[j]]));
	  if (!processed)
	    {
	      sc1[j]=calcMaskIdxAliScore (ali1, umask1, dids1[j], scoring, W); /* calc score sc1 from j to all sequences in alignment 1 */
	      sc2[j]=calcMaskIdxAliScore (ali2, umask2, dids2[j], scoring, W); /* calc score sc2 from j to all sequences in alignment 2 */

	      /* get the alignment with the lower score sc1 or sc2 */
	      if (sc1[j]<sc2[j]) /* score smaller in first alignment */
		{
		  scM[j]=sc1[j];
		  fromwhich[j]=1;
		}
	      else if (sc1[j]>sc2[j]) /* score smaller in second alignment */
		{
		  scM[j]=sc2[j];
		  fromwhich[j]=2;
		}
	      else /* both scores equal */
		{
		  scM[j]=sc1[j];
		  fromwhich[j]=0;
		}
	    }
	}

#ifdef DEBUG
      fprintf (debugfile, "purge2: cross sequence scores...\n");
      for (j=0; j<ndid; j++)
	{
	  processed=((umask1[dids1[j]]) || (umask2[dids2[j]]));
	  if (!processed)
	    fprintf (debugfile, "num: %d SC1: %6.4f SC2: %6.4f SCM: %6.4f fromwhich: %d\n", j, sc1[j], sc2[j], scM[j], fromwhich[j]);
	}
#endif

      /* find sequence with smallest score */
      selM=0.0f;
      selj=0;
      found=0;
      for (j=0; j<ndid; j++)
	{
	  processed=(umask1[dids1[j]] || umask2[dids2[j]]);
	  if (!processed)
	    {
	      if ((!found) || (scM[j]<selM))
		{
		  selM=scM[j];
		  selj=j;
		  found=1;
		}
	    }
	}

#ifdef DEBUG
      fprintf (debugfile, "purge2: sequence with smallest cross score is sequence %d.\n", selj);
#endif

      if (!found) FATAL ("All sequences already disabled.");
      if (umask1[dids1[selj]]) FATAL ("Sequence already disabled.");
      if (umask2[dids2[selj]]) FATAL ("Sequence already disabled.");

      /* disable it from the alignment where the score is less */
      if (fromwhich[selj]==1) 
	{
	  umask1[dids1[selj]]=1; /* disable sequence in first alignment */
	  numdis1++;
#ifdef DEBUG
	  fprintf (debugfile, "purge2: disable sequence %d in 1st alignment.\n", ali1->sids[dids1[selj]]);
#endif
	}
      else if (fromwhich[selj]==2)
	{ 
	  umask2[dids2[selj]]=1; /* disable sequence in second alignment */
	  numdis2++;
#ifdef DEBUG
	  fprintf (debugfile, "purge2: disable sequence %d in 2nd alignment.\n", ali2->sids[dids2[selj]]);
#endif
	}
      else /* equal scores */
	{
	  if (numdis1<numdis2) /* more sequences in alignment 2 disabled */
	    {
	      umask1[dids1[selj]]=1; /* disable sequence in first alignment */
	      numdis1++;
#ifdef DEBUG
	      fprintf (debugfile, "purge2: disable sequence %d in 1st alignment.\n", ali1->sids[dids1[selj]]);
#endif
	    }
	  else /* more sequences in alignment 1 disabled */
	    {
	      umask2[dids2[selj]]=1; /* disable sequence in second alignment */
	      numdis2++;
#ifdef DEBUG
	      fprintf (debugfile, "purge2: disable sequence %d in 2nd alignment.\n", ali2->sids[dids2[selj]]);
#endif
	    }
	}
    }

  /* remove disabled sequences */
  /* first alignment */
  for (i=0; i<num1;)
    {
      if (umask1[i]) /* delete sequence i */
	{
#ifdef DEBUG
	  fprintf (debugfile, "purge2: deleting sequence %d in first alignment.\n", ali1->sids[i]);
#endif
	  FREE1D (ali1->descriptors[i]);
	  FREE1D (ali1->sequences[i]);
	  FREE1D (ali1->sinfo[i]);
	  for (j=i; j<num1-1; j++) /* move all j+1 to j for j=i..n-1 */
	    {
	      ali1->sids[j]=ali1->sids[j+1];
	      ali1->descriptors[j]=ali1->descriptors[j+1];
	      ali1->sequences[j]=ali1->sequences[j+1];
	      ali1->sinfo[j]=ali1->sinfo[j+1];
	      umask1[j]=umask1[j+1];
	    }
	  num1--;
	}
      else i++;
    }
  REALLOC (ali1->sids, int, num1);
  REALLOC (ali1->descriptors, char*, num1);
  REALLOC (ali1->sequences, char*, num1);
  REALLOC (ali1->sinfo, SINFO*, num1);
  ali1->num=num1;
	
  /* second alignment */
  for (i=0; i<num2;)
    {
      if (umask2[i])
	{
#ifdef DEBUG
	  fprintf (debugfile, "purge2: deleting sequence %d in second alignment.\n", ali2->sids[i]);
#endif
	  FREE1D (ali2->descriptors[i]);
	  FREE1D (ali2->sequences[i]);
	  FREE1D (ali2->sinfo[i]);
	  for (j=i; j<num2-1; j++)
	    {
	      ali2->sids[j]=ali2->sids[j+1];
	      ali2->descriptors[j]=ali2->descriptors[j+1];
	      ali2->sequences[j]=ali2->sequences[j+1];
	      ali2->sinfo[j]=ali2->sinfo[j+1];
	      umask2[j]=umask2[j+1];
	    }
	  num2--;
	}
      else i++;
    }
  REALLOC (ali2->sids, int, num2);
  REALLOC (ali2->descriptors, char*, num2);
  REALLOC (ali2->sequences, char*, num2);
  REALLOC (ali2->sinfo, SINFO*, num2);
  ali2->num=num2;

  /* free memory */
  FREE1D (umask1);
  FREE1D (umask2);
  FREE1D (fromwhich);
  FREE1D (sc1);
  FREE1D (sc2);
  FREE1D (scM);

#ifdef DEBUG
  fprintf (debugfile, "purge1: alignments after erasing duplicated sequences:\n");
  fprintf (debugfile, "ali1: #seqs=%d, len=%d\n", ali1->num, ali1->len);
  for (i=0; i<ali1->num; i++) fprintf (debugfile, "%s: %s\n", ali1->descriptors[i], ali1->sequences[i]);
  fprintf (debugfile, "ali2: #seqs=%d, len=%d\n", ali2->num, ali2->len);
  for (i=0; i<ali2->num; i++) fprintf (debugfile, "%s: %s\n", ali2->descriptors[i], ali2->sequences[i]);
#endif
}
#endif
