#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "global.h"
#include "options.h"
#include "macros.h"
#include "scores.h"
#include "ali.h"

#ifdef DEBUG
extern FILE *debugfile;
#endif

/***********************************************************
calculate sequence score
************************************************************/
static float getScore (char c1, char c2, SCORING *scoring, SINFO *s1, SINFO *s2)
{
  float seqscore, structscore, score;
  int cde1=c1-OFFSET;
  int cde2=c2-OFFSET;
  BOOL struc=scoring->usestruc;
  float psi=scoring->psi;
  float **M=(float**)scoring->M;

  if ((cde1>=0) && (cde1<26) && (cde2>=0) && (cde2<26)) seqscore=M[cde1][cde2]; /* ignore unknown characters */
  else seqscore=0.0f;
  
  if ((struc) && (s1) && (s2))
    {
#if SCORESTRAL /* STRAL like scoring */
      seqscore=sqrt (s1->unpaired*s2->unpaired)*seqscore;
      structscore=Options.Psi*(sqrt (s1->dnpaired*s2->dnpaired)+sqrt (s1->uppaired*s2->uppaired));
#else 
      seqscore=psi*(seqscore);
      structscore=(1.0f-psi)*(sqrt (s1->dnpaired*s2->dnpaired)+sqrt (s1->uppaired*s2->uppaired)+sqrt (s1->unpaired*s2->unpaired));
#endif
      score=structscore+seqscore;
    }
  else
    {
      score=seqscore;
    }
  return (score);
}

/*******************************************************
get the score of an alignment
*******************************************************/
float calcAliScore (ALI *ali, SCORING *scoring, float *W)
{
  int i, j, k;
  int g1=0, g2=0;
  float gs1=0.0f, gs2=0.0f;
  float w1, w2;
  char c1, c2;
  float score;
  float norm;
  float gapweight;
  int *posbeg=NULL, *posend=NULL;
  SINFO *sinfo1, *sinfo2;
  
  /* get data */
  int num=ali->num;
  int len=ali->len;
  float go=scoring->go;
  float ge=scoring->ge;
  BOOL tgw=scoring->termgapweight;

  if (!tgw)
    {
      MAKE1DINT (posbeg, num);
      MAKE1DINT (posend, num);

      /* get positions of terminal gaps */
      for (i=0; i<num; i++)
	{
	  for (k=0; k<len; k++)
	    {
	      if (ali->sequences[i][k]!=GAPCHAR)
		{
		  posbeg[i]=k;
		  break;
		}
	    }

	  for (k=len-1; k>=0; k--)
	    {
	      if (ali->sequences[i][k]!=GAPCHAR)
		{
		  posend[i]=k;
		  break;
		}
	    }
	}  
    }

  score=0.0f;
  norm=0.0f;
  /* iterate over all pairs of sequences */
  for (i=0; i<num; i++) for (j=0; j<i; j++)
    {
      /* get weights */
      if (W)
	{
	  w1=W[ali->sids[i]];
	  w2=W[ali->sids[j]];
	}
      else
	{
	  w1=1.0f;
	  w2=1.0f;
	}
      norm+=(w1*w2);
      
      /* iterate over length of alignment */
      for (k=0; k<len; k++)
	{
	  /* get character */
	  c1=ali->sequences[i][k];
	  c2=ali->sequences[j][k];

	  /* get structural information */
	  if (ali->sinfo)
	    {
	      sinfo1=ali->sinfo[i]+k;
	      sinfo2=ali->sinfo[j]+k;
	    }
	  else
	    {
	      sinfo1=NULL;
	      sinfo2=NULL;
	    }

	  gapweight=1.0f;

	  if (!tgw)
	    {
	      if ((k<posbeg[i]) || (k>posend[i])) gapweight=0.0f;
	      if ((k<posbeg[j]) || (k>posend[j])) gapweight=0.0f;
	    }

	  /* gap */
	  if (c1==GAPCHAR)
	    {
	      g1=1;
	      /* new gap or extending existing one */
	      if ((k==0) || (ali->sequences[i][k-1]!=GAPCHAR)) gs1=go; else gs1=ge;
	    }
	  else g1=0;

	  /* gap */
	  if (c2==GAPCHAR)
	    {
	      g2=1;
	      /* new gap or extending existing one */
	      if ((k==0) || (ali->sequences[j][k-1]!=GAPCHAR)) gs2=go; else gs2=ge;
	    }
	  else g2=0;

	  /* get score at current position */
	  if ((!g1) && (!g2)) score+=(w1*w2*getScore (c1, c2, scoring, sinfo1, sinfo2)); /* ignore unknown characters */
	  else if ((g1) && (!g2)) score-=(w1*w2*gs1*gapweight);
	  else if ((!g1) && (g2)) score-=(w1*w2*gs2*gapweight);
	}
    }

  if (!tgw)
    {
      FREE1D (posbeg);
      FREE1D (posend);
    }

  /* normalize with number of sequence pairs */
  if (norm>0.0f) ali->score=score/norm;
  else ali->score=0.0f;

  return (ali->score);
}

/*******************************************************
get the alignment score for an alignment with disabled
sequences
*******************************************************/
float calcMaskAliScore (ALI *ali, char *umask, SCORING *scoring, float *W)
{
  int i, j, k;
  int g1=0, g2=0;
  float gs1=0.0f, gs2=0.0f;
  float w1, w2;
  char c1, c2;
  float score;
  float gapweight;
  float norm;
  int *posbeg=NULL, *posend=NULL;
  SINFO *sinfo1, *sinfo2;

  /* get data */
  int num=ali->num;
  int len=ali->len;
  float go=scoring->go;
  float ge=scoring->ge;
  BOOL tgw=scoring->termgapweight;

  if (!tgw)
    {
      MAKE1DINT (posbeg, num);
      MAKE1DINT (posend, num);

      /* get positions of terminal gaps */
      for (i=0; i<num; i++)
	{
	  for (k=0; k<len; k++)
	    {
	      if (ali->sequences[i][k]!=GAPCHAR)
		{
		  posbeg[i]=k;
		  break;
		}
	    }

	  for (k=len-1; k>=0; k--)
	    {
	      if (ali->sequences[i][k]!=GAPCHAR)
		{
		  posend[i]=k;
		  break;
		}
	    }
	}  
    }

  score=0.0f;
  norm=0.0f;
  /* iterate over all pairs of sequences */
  for (i=0; i<num; i++) for (j=0; j<i; j++)
    {
      if ((!umask[i]) && (!umask[j]))
	{
	  /* get weights */
	  if (W)
	    {
	      w1=W[ali->sids[i]];
	      w2=W[ali->sids[j]];
	    }
	  else
	    {
	      w1=1.0f;
	      w2=1.0f;
	    }
	  norm+=(w1*w2);
	  			
	  /* iterate over length of alignment */
	  for (k=0; k<len; k++)
	    {
	      /* get character */
	      c1=ali->sequences[i][k];
	      c2=ali->sequences[j][k];

	      /* get structural information */
	      sinfo1=ali->sinfo[i]+k;
	      sinfo2=ali->sinfo[j]+k;

	      gapweight=1.0f;

	      if (!tgw)
		{
		  if ((k<posbeg[i]) || (k>posend[i])) gapweight=0.0f;
		  if ((k<posbeg[j]) || (k>posend[j])) gapweight=0.0f;
		}

	      /* gap */
	      if (c1==GAPCHAR)
		{
		  g1=1;
		  /* new gap or extending existing one */
		  if ((k==0) || (ali->sequences[i][k-1]!=GAPCHAR)) gs1=go; else gs1=ge;
		}
	      else g1=0;

	      /* gap */
	      if (c2==GAPCHAR)
		{
		  g2=1;
		  /* new gap or extending existing one */
		  if ((k==0) || (ali->sequences[j][k-1]!=GAPCHAR)) gs2=go; else gs2=ge;
		}
	      else g2=0;

	      /* get score at current position */
	      if ((!g1) && (!g2)) score+=(w1*w2*getScore (c1, c2, scoring, sinfo1, sinfo2)); /* ignore unknown characters */
	      else if ((g1) && (!g2)) score-=(w1*w2*gs1*gapweight);
	      else if ((!g1) && (g2)) score-=(w1*w2*gs2*gapweight);
	    }
	}
    }
  if (!tgw)
    {
      FREE1D (posbeg);
      FREE1D (posend);
    }

  /* normalize with number of sequence pairs */
  if (norm>0.0f) score/=norm;
  else score=0.0f;

  return (score);
}

/*******************************************************
get the pairwise score of an alignment for all pairs
with a fixed sequence and a mask of disabled sequences
comparisons are (0:idx,1:idx,idx:idx,...,n:idx),
sequence is also compared to itself
*******************************************************/
float calcMaskIdxAliScore (ALI *ali, char *umask, int idx, SCORING *scoring, float *W)
{
  int i, k;
  int g1=0, g2=0;
  float gs1=0.0f, gs2=0.0f;
  float w1, w2;
  char c1, c2;
  float score;
  float norm;
  float gapweight;
  int *posbeg=NULL, *posend=NULL;
  SINFO *sinfo1, *sinfo2;

  /* get data */
  int num=ali->num;
  int len=ali->len;
  float go=scoring->go;
  float ge=scoring->ge;
  BOOL tgw=scoring->termgapweight;

  if (!tgw)
    {
      MAKE1DINT (posbeg, num);
      MAKE1DINT (posend, num);

      /* get positions of terminal gaps */
      for (i=0; i<num; i++)
	{
	  for (k=0; k<len; k++)
	    {
	      if (ali->sequences[i][k]!=GAPCHAR)
		{
		  posbeg[i]=k;
		  break;
		}
	    }

	  for (k=len-1; k>=0; k--)
	    {
	      if (ali->sequences[i][k]!=GAPCHAR)
		{
		  posend[i]=k;
		  break;
		}
	    }
	}  
    }

  score=0.0f;
  norm=0.0f;
  /* iterate over all pairs of sequences */
  for (i=0; i<num; i++)
    {
      if ((!umask[i]) && (!umask[idx]))
	{
	  /* get weights */
	  if (W)
	    {
	      w1=W[ali->sids[i]];
	      w2=W[ali->sids[idx]];
	    }
	  else
	    {
	      w1=1.0f;
	      w2=1.0f;
	    }
	  norm+=(w1*w2);
	  
	  /* iterate over length of alignment */
	  for (k=0; k<len; k++)
	    {
	      /* get character */
	      c1=ali->sequences[i]  [k];
	      c2=ali->sequences[idx][k];

	      /* get structural information */
	      sinfo1=ali->sinfo[i]  +k;
	      sinfo2=ali->sinfo[idx]+k;

	      gapweight=1.0f;

	      if (!tgw)
		{
		  if ((k<posbeg[i]) || (k>posend[i])) gapweight=0.0f;
		  if ((k<posbeg[idx]) || (k>posend[idx])) gapweight=0.0f;
		}

	      /* gap */
	      if (c1==GAPCHAR)
		{
		  g1=1;
		  /* new gap or extending existing one */
		  if ((k==0) || (ali->sequences[i][k-1]!=GAPCHAR)) gs1=go; else gs1=ge;
		}
	      else g1=0;

	      /* gap */
	      if (c2==GAPCHAR)
		{
		  g2=1;
		  /* new gap or extending existing one */
		  if ((k==0) || (ali->sequences[idx][k-1]!=GAPCHAR)) gs2=go; else gs2=ge;
		}
	      else g2=0;

	      /* get score at current position */
	      if ((!g1) && (!g2)) score+=(w1*w2*getScore (c1, c2, scoring, sinfo1, sinfo2)); /* ignore unknown charcters */
	      else if ((g1) && (!g2)) score-=(w1*w2*gs1*gapweight);
	      else if ((!g1) && (g2)) score-=(w1*w2*gs2*gapweight);
	    }
	}
    }
  if (!tgw)
    {
      FREE1D (posbeg);
      FREE1D (posend);
    }

  /* normalize with number of sequence pairs */
  if (norm>0.0f) score/=norm;
  else score=0.0f;

  return (score);
}

/*******************************************************
calculate pairwise score of two alignments
*******************************************************/
float **calcPairwiseScore (ALI *ali1, ALI *ali2, SCORING *scoring, float *W)
{
  int i1,i2;
  int j1,j2;
  int g1=0, g2=0;
  char c1, c2;
  float w1, w2;
  float norm;
  float gs1=0.0f, gs2=0.0f;
  float gapweight1, gapweight2;
  float **S=NULL;
  int *posbeg1=NULL, *posend1=NULL;
  int *posbeg2=NULL, *posend2=NULL;
  SINFO *sinfo1, *sinfo2;

  /* get data */
  int num1=ali1->num;
  int num2=ali2->num;
  int len1=ali1->len;
  int len2=ali2->len;
  float go=scoring->go;
  float ge=scoring->ge;
  BOOL tgw=scoring->termgapweight;

  if (!tgw)
    {
      MAKE1DINT (posbeg1, num1);
      MAKE1DINT (posend1, num1);
      MAKE1DINT (posbeg2, num2);
      MAKE1DINT (posend2, num2);

      /* get positions of terminal gaps */
      /* first alignment */
      for (i1=0; i1<num1; i1++)
	{
	  for (j1=0; j1<len1; j1++)
	    {
	      if (ali1->sequences[i1][j1]!=GAPCHAR)
		{
		  posbeg1[i1]=j1;
		  break;
		}
	    }
	  for (j1=len1-1; j1>=0; j1--)
	    {
	      if (ali1->sequences[i1][j1]!=GAPCHAR)
		{
		  posend1[i1]=j1;
		  break;
		}
	    }
	}  
      /* second alignment */
      for (i2=0; i2<num2; i2++)
	{
	  for (j2=0; j2<len2; j2++)
	    {
	      if (ali2->sequences[i2][j2]!=GAPCHAR)
		{
		  posbeg2[i2]=j2;
		  break;
		}
	    }
	  for (j2=len2-1; j2>=0; j2--)
	    {
	      if (ali2->sequences[i2][j2]!=GAPCHAR)
		{
		  posend2[i2]=j2;
		  break;
		}
	    }
	}  
    }

  /* allocate memory for matrix */
  MAKE2DFLOAT (S, len1, len2, i1);

  norm=0.0f;
  /* compare sequences in first alignment with sequences in second alignment */
  for (i1=0; i1<num1; i1++) for (i2=0; i2<num2; i2++)
    {
      /* get weights */
      if (W)
	{
	  w1=W[ali1->sids[i1]];
	  w2=W[ali2->sids[i2]];
	}
      else
	{
	  w1=1.0f;
	  w2=1.0f;
	}
      norm+=(w1*w2);

      /* iterate over loci */
      for (j1=0; j1<len1; j1++) for (j2=0; j2<len2; j2++)
	{
	  /* get character */
	  c1=ali1->sequences[i1][j1];
	  c2=ali2->sequences[i2][j2];

	  /* get structural information */
	  sinfo1=ali1->sinfo[i1]+j1;
	  sinfo2=ali2->sinfo[i2]+j2;

	  gapweight1=1.0f;
	  gapweight2=1.0f;

	  if (!tgw)
	    {
	      if ((j1<posbeg1[i1]) || (j1>posend1[i1])) gapweight1=0.0f;
	      if ((j2<posbeg2[i2]) || (j2>posend2[i2])) gapweight2=0.0f;
	    }

	  /* check for gaps */
	  if (ali1->sequences[i1][j1]==GAPCHAR)
	    {
	      g1=1;
	      /* new gap or extending old one */
	      if ((j1==0) || (ali1->sequences[i1][j1-1]!=GAPCHAR)) gs1=go; else gs1=ge;
	    }
	  else g1=0;
	  /* check for gaps */
	  if (ali2->sequences[i2][j2]==GAPCHAR)
	    {
	      g2=1;
	      /* new gap or extending old one */
	      if ((j2==0) || (ali2->sequences[i2][j2-1]!=GAPCHAR)) gs2=go; else gs2=ge;
	    }
	  else g2=0;

	  /* get score for this position */
	  if ((!g1) && (!g2)) S[j1][j2]+=(w1*w2*getScore (c1, c2, scoring, sinfo1, sinfo2)); /* ignore unknown characters */
	  else if ((g1) && (!g2)) S[j1][j2]-=(w1*w2*gs1*gapweight1);
	  else if ((!g1) && (g2)) S[j1][j2]-=(w1*w2*gs2*gapweight2);
	}
    }

  if (!tgw)
    {
      FREE1D (posbeg1);
      FREE1D (posend1);
      FREE1D (posbeg2);
      FREE1D (posend2);
    }

#ifdef DEBUG
  fprintf (debugfile, "calcPairwiseScore (raw)\n");
  for (j1=0; j1<len1; j1++)
    {
      for (j2=0; j2<len2; j2++) fprintf (debugfile, "%6.3f ", S[j1][j2]);
      fprintf (debugfile, "\n");
    }
#endif

  /* normalize score with weighted number of sequences compared */
  if (norm>0.0f) for (j1=0; j1<len1; j1++) for (j2=0; j2<len2; j2++) S[j1][j2]/=norm;

#ifdef DEBUG
  fprintf (debugfile, "normalization factor is 1/%6.3f ", norm);
#endif

#ifdef DEBUG
  fprintf (debugfile, "calcPairwiseScore\n");
  for (j1=0; j1<len1; j1++)
    {
      for (j2=0; j2<len2; j2++) fprintf (debugfile, "%6.3f ", S[j1][j2]);
      fprintf (debugfile, "\n");
    }
#endif

  return (S);
}
