#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "global.h"
#include "options.h"
#include "seqio.h"
#include "macros.h"
#include "ali.h"

#ifdef DEBUG
extern FILE *debugfile;
#endif

extern OPTION Options;
extern SEQ Seqs;

/********************************************************
return code of residue
********************************************************/
static int res2Code (char res, SEQTYPE seqtype)
{
  char c=tolower (res);

  if (seqtype==SEQ_DNA)
    {
      switch (c)
	{
	case 'a':
	case 'c':
	case 'g':
	case 't': return ((int)c-OFFSET);
	default: return (-1);
	}
    }
  else if (seqtype==SEQ_RNA)
    {
      switch (c)
	{
	case 'a':
	case 'c':
	case 'g':
	case 'u': return ((int)c-OFFSET);
	default: return (-1);
	}
    }
  else if (seqtype==SEQ_PROTEIN)
    {
      switch (c)
	{
	case 'a':
	case 'b':
	case 'c':
	case 'd':
	case 'e':
	case 'f':
	case 'g':
	case 'h':
	case 'i':
	case 'k':
	case 'l':
	case 'm':
	case 'n':
	case 'p':
	case 'q':
	case 'r':
	case 's':
	case 't':
	case 'v':
	case 'w':
	case 'x':
	case 'y':
	case 'z': return ((int)c-OFFSET);
	default: return (-1);
	}
    }
  else FATAL ("Unknown sequence type.");
  return (-1);
}

/********************************************************
determine sequence type and convert to lower case
*********************************************************/
static void detSeqType (void)
{
  int i, j;
  int numres=0, numdna=0, numrna=0, numprotein=0;
  
  /* get residues */
  for (i=0; i<Seqs.num; i++)
    {
      for (j=0; Seqs.sequences[i][j]!=EOS; j++)
	{
	  Seqs.sequences[i][j]=tolower (Seqs.sequences[i][j]);
	  if (Seqs.sequences[i][j]!=GAPCHAR)
	    {
	      numres++;
	      if (res2Code (Seqs.sequences[i][j], SEQ_DNA)>=0) numdna++;
	      if (res2Code (Seqs.sequences[i][j], SEQ_RNA)>=0) numrna++;
	      if (res2Code (Seqs.sequences[i][j], SEQ_PROTEIN)>=0) numprotein++;
	    }
	}
    }

  /* determine sequence type */
  if (Options.Seqtype==SEQ_DNA)
    {
      if (numdna<numres) printf ("%d unknown letters.\n", numres-numdna);
      Seqs.type=SEQ_DNA;
    }
  else if (Options.Seqtype==SEQ_RNA)
    {
      if (numrna<numres) printf ("%d unknown letters.\n", numres-numrna);
      Seqs.type=SEQ_RNA;
    }
  else if (Options.Seqtype==SEQ_PROTEIN)
    {
      if (numprotein<numres) printf ("%d unknown letters.\n", numres-numprotein);
      Seqs.type=SEQ_PROTEIN;
    }
  else /* autodetect sequence type */
    {
      if (((double)numdna/(double)numres)>0.95) Seqs.type=SEQ_DNA;
      else if (((double)numrna/(double)numres)>0.95) Seqs.type=SEQ_RNA;
      else if (((double)numprotein/(double)numres)>0.95) Seqs.type=SEQ_PROTEIN;
      else FATAL ("Cannot determine sequence type. Please specify manually.");

      if (Options.Verbosity) 
	printf ("Setting sequence type to %s.\n", (Seqs.type==SEQ_DNA ? "DNA" : 
						   (Seqs.type==SEQ_RNA ? "RNA" : 
						    (Seqs.type==SEQ_PROTEIN ? "PROTEIN" : "UNKNOWN"))));
    }
}

/*******************************************************
read sequences from FASTA file
*******************************************************/
void loadSequences (char *filename)
{
  char **dess=NULL;
  char **seqs=NULL;
  int numseqs;
  int deslen, seqlen;
  int c;
  int cpos;
  FILE *in=NULL;

  /* open file */
  if ((in=fopen (filename, "r"))==0) FATAL ("Cannot open input file!");

  numseqs=0;
  /* read sequences from file */
  do
    {
      c=fgetc (in);
      /* remark */
      if ((c=='#') || (c=='!'))
	{
	  while ((!feof (in)) && (c!='\n')) c=fgetc (in); /* read whole line */
	}
      /* FASTA descriptor */
      else if (c=='>')
	{
	  /* ready to start a new sequence */
	  deslen=0;
	  REALLOC (dess, char*, numseqs+1);
	  REALLOC (seqs, char*, numseqs+1);
	  dess[numseqs]=NULL;
	  seqs[numseqs]=NULL;
	  numseqs++;
	  c=fgetc (in);
	  /* read whole line */
	  while ((!feof (in)) && (c!='\n'))
	    {
	      if (c!=' ')
		{
		  REALLOC (dess[numseqs-1], char, deslen+1);
		  dess[numseqs-1][deslen++]=(char)c;
		  c=fgetc (in);
		}
	      else if ((c==' ') && (deslen>0))
		{
		  while ((!feof (in)) && (c!='\n')) c=fgetc (in);
		}
	      else /* c==' ' && len==0 */
		{
		  c=fgetc (in);
		}
	    }

	  if (deslen)
	    {
	      /* terminate descriptor */
	      REALLOC (dess[numseqs-1], char, deslen+1);
	      dess[numseqs-1][deslen]=EOS;
	    }

	  /* now the sequence has to be read */
	  /* read until new fasta descriptor or remark */
	  c=fgetc (in);
	  seqlen=0;
	  while ((!feof (in)) && (c!='>') && (c!='#'))
	    {
	      if (((c>='a') && (c<='z')) || ((c>='A') && (c<='Z'))) /* letter */
		{
		  c=tolower (c);
		  REALLOC (seqs[numseqs-1], char, seqlen+1);
		  seqs[numseqs-1][seqlen]=(char)c;
		  seqlen++;
		}
	      c=fgetc (in);
	    }
	  ungetc (c, in); /* put last char back to stream buffer */

	  if (seqlen)
	    {
	      /* terminate sequence */
	      REALLOC (seqs[numseqs-1], char, seqlen+1);
	      seqs[numseqs-1][seqlen]=EOS;
	    }
	  else
	    {
	      /* no sequence to current fasta identifier -> free memory */
	      free (dess[numseqs-1]);
	      free (seqs[numseqs-1]);
	      numseqs--;
	      REALLOC (dess, char*, numseqs);
	      REALLOC (seqs, char*, numseqs);
	    }
	}
    }
  while (!feof (in));

  /* close file handle */
  fclose (in);

  /* check number of sequences */
  if (numseqs<3) FATAL ("At least three sequences are needed for a multiple sequence alignment.");

  /* set sequence record entries */
  Seqs.descriptors=dess;
  Seqs.sequences=seqs;
  Seqs.num=numseqs;
  Seqs.type=SEQ_AUTO;

  /* detect sequence type */
  detSeqType ();

  /* output FASTA descriptors, if desired */
  if (Options.Verbosity>1)
    {
      cpos=0;
      for (c=0; c<Seqs.num; c++)
	{
	  cpos+=(strlen (Seqs.descriptors[c])+3);
	  if (cpos>=80)
	    {
	      printf ("\n");
	      cpos=0;
	    }
	  printf (" [%s]", Seqs.descriptors[c]);
	}
      if (cpos) printf ("\n");
    }

#ifdef DEBUG
  fprintf (debugfile, "loadSequences: %d sequences loaded\n", numseqs);
  for (c=0; c<numseqs; c++) fprintf (debugfile, "%s\n%s\n", dess[c], seqs[c]);
#endif
}

/**********************************************************
write an alignment in CLUSTAL format to file
***********************************************************/
void saveAliClustal (char *filename, ALI *ali)
{
  int i, j, k;
  char *cons=NULL;
  int err;
  static char buffer[LW+1];
  FILE *out=NULL;

  /* get data */
  int num=ali->num;
  int len=ali->len;

  MALLOC (cons, char, len+1);
  cons[len]=EOS;
  /* calculate consensus of sequences */
  for (i=0; i<len; i++)
    {
      err=0;
      for (j=0; j<num; j++) for (k=0; k<j; k++) if (ali->sequences[j][i]!=ali->sequences[k][i]) err=1;
      if (!err) cons[i]='*';
      else cons[i]=' ';
    }

  /* open file */
  if ((out=fopen (filename, "w"))==0) FATAL ("Cannot create output file.");
  
  /* write header */
  fprintf (out, "CLUSTAL W (1.83) multiple sequence alignment\n\n\n");
      
  /* write sequences to file*/
  for (i=0; i<len; i+=LW)
    {
      /* write sequences */
      for (j=0; j<num; ++j)
	{
	  strncpy (buffer, ali->sequences[j]+i, LW);
	  buffer[LW]='\0';
	  for (k=0; buffer[k]!=EOS; k++) buffer[k]=toupper (buffer[k]); /* convert to upper-case */
	  fprintf (out, "%-26.20s%s\n", ali->descriptors[j], buffer);
	}
      /* write consensus */
      strncpy (buffer, cons+i, LW);
      fprintf (out, "%-26.20s%s\n", " ", buffer);
      fprintf (out, "\n");
    }
  fclose (out);

  /* free memory */
  FREE1D (cons);
}

/**********************************************************
write an alignment in MSF format to file
***********************************************************/
void saveAliMsf (char *filename, ALI *ali)
{
  int i, j, p;
  FILE *out=NULL;

  /* get data */
  int num=ali->num;
  int len=ali->len;

  /* open file */
  if ((out=fopen (filename, "w"))==0) FATAL ("Cannot create output file.");
  
  /* write header */
  fprintf (out, "PileUp\n\n");
  fprintf (out, " MSF: %4d  Type: %c  Check: 1%03d   ..\n\n", len, (((Seqs.type==SEQ_DNA) || (Seqs.type==SEQ_RNA)) ? 'N' : 'P'), ali->id);
  for (i=0; i<ali->num; i++)
    fprintf (out, " Name: %-20s  Len: %4d  Check: 1%03d  Weight: %5.1f\n", ali->descriptors[i], len, ali->sids[i], 1.0);
  fprintf (out, "\n");
  fprintf (out, "//\n\n\n");
  
  /* write content */
  for (p=0; p<len; p+=50)
    {
      for (i=0; i<num; i++)
	{
	  fprintf (out, "%-20s", ali->descriptors[i]);
	  for (j=p; (j<p+50) && (j<len); j++)
	    {
	      if (j%10==0) fprintf (out, " "); /* insert blank every 10th residue */
	      fprintf (out, "%c", ali->sequences[i][j]==GAPCHAR ? '.' : ali->sequences[i][j]);
	    }
	  fprintf (out, "\n");
	}
      fprintf (out, "\n\n"); /* block separator */
    }
  fclose (out);
}

/**********************************************************
load an alignment in CLUSTAL format from file
***********************************************************/
ALI *loadAliClustal (char *filename)
{
  char **des=NULL;
  char **seq=NULL;
  char *fwdspace, *bckspace;
  int len, spacelen, deslen, seqlen, oldlen;
  int block; /* sequence block in file */
  int idx; /* sequence index */
  int numseqs; /* number of sequences */
  FILE *in=NULL;
  static char line[BUFLEN];
  ALI *ali=NULL;

  /* open file */
  if ((in=fopen (filename, "r"))==0) FATAL ("Cannot open alignment file.");

  if (!feof (in)) fgets (line, BUFLEN, in); /* remark line */

  idx=0;
  block=0;
  numseqs=0;
  /* read file entries */
  while (!feof (in))
    {
      fgets (line, BUFLEN, in);
      len=(int)strlen (line);

      /* important content */
      if ((!feof (in)) && (len>0) && (line[0]!=' ') && (line[0]!='\n'))
	{
	  /* remove trailing newline */
	  line[--len]='\0';
	  /* remove trailing whitespaces */
	  while (line[len-1]==' ') line[--len]='\0';

	  fwdspace=strchr (line, ' '); /* first occurence of whitespace */
	  bckspace=strrchr (line, ' '); /* last occurence of whitespace */
	  spacelen=(int)(bckspace-fwdspace); /* length of spacefield */
	  deslen=(int)(fwdspace-line); /* length of descriptor */
	  seqlen=(int)(len-spacelen-deslen); /* length of sequence in this block  */
	  if ((deslen) && (seqlen))
	    {
	      if (!block) ++numseqs; /* if first block -> increase number of sequences */
	      if (!block) /* if first block -> re-allocate memory for descriptors and sequences */
		{
		  REALLOC (des, char*, numseqs);
		  REALLOC (seq, char*, numseqs);
		}
	      if (!block) /* first block */
		{
		  MALLOC (des[idx], char, deslen+1);
		  strncpy (des[idx], line, deslen);
		  des[idx][deslen]='\0';
		  MALLOC (seq[idx], char, seqlen+1);
		  strncpy (seq[idx], bckspace+1, seqlen);
		}
	      else /* further block */
		{
		  oldlen=(int)strlen (seq[idx]);
		  REALLOC (seq[idx], char, oldlen+seqlen+1);
		  strncat (seq[idx], bckspace+1, seqlen);
		}
	      ++idx;
	    }
	}
      else /* newline or first character is whitespace -> block separator */
	{
	  if (idx) ++block;
	  idx=0; /* reset index */
	}
    }
  fclose (in);

  if (numseqs==0) FATAL ("No sequences found.");

  /* convert to alignment structure */
  MALLOC (ali, ALI, 1);
  ali->num=numseqs;
  ali->len=(int)strlen (seq[0]);
  ali->descriptors=des;
  ali->sequences=seq;

  /* convert alignment to sequences */
  Seqs=ali2Seq (ali);

  /* detect sequence type */
  detSeqType ();
 
  return (ali);
}
