#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "tree.h"
#include "macros.h"

#if DEBUG
extern FILE *debugfile;
#endif

/*******************************************************
calculate edge weight of every edge in tree
********************************************************/
static void calcEdgeWeight (EDGE *edge)
{
  NODE *tmpnode;
  int i;

  /* calculate weight of this edge */
  edge->weight=(edge->length)/(float)(edge->subtreesize);

  /* get node edge points to */
  tmpnode=edge->to;

  /* invoke recursion of every edge leaving node this edge points to */
  for (i=0; i<tmpnode->outnum; i++) calcEdgeWeight (tmpnode->out[i]);
}

/********************************************************
traversing tree from root to all leaves and calculate
node weights
*********************************************************/
static void calcNodeWeight (NODE *node, float weight)
{
  NODE *tmpnode;
  EDGE *tmpedge;
  int i;
  float tmpweight;

  if (!node->outnum) node->weight=weight;
  else
    {
      node->weight=0.0f; /* all non-terminal nodes receive a weight of zero */
      for (i=0; i<node->outnum; i++)
	{
	  tmpedge=node->out[i];
	  tmpnode=tmpedge->to;
	  tmpweight=tmpedge->weight;
	  /* recurisvely calculate weight of subsequent nodes */
	  calcNodeWeight (tmpnode, weight+tmpweight);
	}
    }
}

/*******************************************************
calculate number of terminal nodes of sub-tree for 
each edge
********************************************************/
static void calcNumTerm (EDGE *edge)
{
  NODE *tmpnode;
  int i;
  int sum=0;

  /* get node edge points to */
  tmpnode=edge->to;

  /* node has no leaving edges */
  if (!tmpnode->outnum)
    {
      edge->subtreesize=1;
    }
  /* node has leaving edges */
  else
    {
      /* recursively calculate number of terminal nodes for all edges leaving node this edge points to */
      for (i=0; i<tmpnode->outnum; i++)
	{
	  calcNumTerm (tmpnode->out[i]);
	  sum+=tmpnode->out[i]->subtreesize;
	}
      edge->subtreesize=sum;
    }
}

/*******************************************************
extracts node weights from tree data type to array of
floats
********************************************************/
static void extractWeights (NODE *node, float *weights)
{
  int i;

  for (i=0; i<node->outnum; i++) extractWeights (node->out[i]->to, weights);
  if (node->id>=0) weights[node->id]=node->weight; /* store node weights of leafs to matrix */
}

/********************************************************
calculate node weights
*********************************************************/
void calcNodeWeights (NODE *root, float *weights)
{
  int i;

  /* first, calculate the size of the sub-tree of each node */
  for (i=0; i<root->outnum; i++) calcNumTerm (root->out[i]);

  /* calculate the edge weights for every edge in the tree */
  for (i=0; i<root->outnum; i++) calcEdgeWeight (root->out[i]);

  /* now calculate weight for every terminal node */
  calcNodeWeight (root, 0.0f);

  /* extract weights from tree data structure to array */
  extractWeights (root, weights);
}

/********************************************************
create a new node
*********************************************************/
NODE *NODEmake (int id)
{
  NODE *node=NULL;

  MALLOC (node, NODE, 1);
  node->id=id;

  return (node);
}

/********************************************************
create a new edge
*********************************************************/
EDGE *EDGEmake (void)
{
  EDGE *edge=NULL;

  MALLOC (edge, EDGE, 1);

  return (edge);
}

/********************************************************
add a given edge between two nodes
*********************************************************/
void NODEaddedge (NODE *nodefrom, NODE *nodeto, EDGE *edge)
{
  edge->from=nodefrom;
  edge->to=nodeto;

  REALLOC (nodefrom->out, EDGE*, nodefrom->outnum+1);
  nodefrom->out[nodefrom->outnum]=edge;
  (nodefrom->outnum)++;

  nodeto->in=edge;
}

/********************************************************
deletes an edge between two nodes
*********************************************************/
int NODEdeledge (NODE *nodefrom, NODE *nodeto)
{
  int i, j;
  EDGE *tmpedge=NULL;

  for (i=0; i<nodefrom->outnum;)
    {
      tmpedge=nodefrom->out[i];
      if (tmpedge->to==nodeto)
	{
	  for (j=i; j<nodefrom->outnum-1; j++) nodefrom->out[j]=nodefrom->out[j+1];
	  REALLOC (nodefrom->out, EDGE*, nodefrom->outnum-1);
	  (nodefrom->outnum)--;
	  nodeto->in=NULL;
	  FREE (tmpedge);
	  return (1);
	}
      else i++;
    }
  return (0);
}

/********************************************************
set weight of an edge between two nodes
*********************************************************/
int NODEsetedgelength (NODE *node1, NODE *node2, float length)
{
  int i;
  EDGE *tmpedge=NULL;

  for (i=0; i<node1->outnum; i++)
    {
      tmpedge=node1->out[i];
      if (tmpedge->to==node2)	
	{
	  tmpedge->length=length;
	  return (1);
	}
    }
  return (0);
}

/*******************************************************
recursively frees tree architecture given by root
********************************************************/
void TREEfree (NODE *node)
{
  int i;

  for (i=0; i<node->outnum; i++) TREEfree (node->out[i]->to);
  if (node->in) FREE (node->in); /* root has no incomming edge */
  FREE (node);
}

/*******************************************************
print a tree data-structure
********************************************************/
#if DEBUG
void printTree (NODE *node)
{
  int i;
  fprintf (debugfile, "nid:%d ", node->id);
  fprintf (debugfile, "nweight:%.2f ", node->weight);
  fprintf (debugfile, " (");
  for (i=0; i<node->outnum; i++)
    {
      fprintf (debugfile, "elength:%.2f ", node->out[i]->length);
      fprintf (debugfile, "eweight:%.2f ", node->out[i]->weight);
      fprintf (debugfile, "st-size:%d ", node->out[i]->subtreesize);
      printTree (node->out[i]->to);
    }
  fprintf (debugfile, ")\n");
}
#endif
