#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "global.h"
#include "options.h"
#include "macros.h"
#include "weights.h"
#include "njoin.h"
#include "tree.h"

#ifdef DEBUG
extern FILE *debugfile;
#endif

extern OPTION Options;

/********************************************************
calculate sequence weights for sequences in given
alignment
********************************************************/
void calcWeights (float *W, float **D, ALI *ali1, ALI *ali2, ALI *ali3)
{
  int i;
  float l1, l2, l3;
  float sum;

  NODE *root, *tree1, *tree2, *tree3;
  EDGE *edge1, *edge2, *edge3;

  /* get data */
  int num1=ali1->num;
  int num2=ali2->num;
  int num3=ali3->num;
  int num=num1+num2+num3;
  
  if (!Options.Seqweight)
    {
      /* same weight for all sequences in the alignment */
      for (i=0; i<num1; i++) W[ali1->sids[i]]=1.0f;
      for (i=0; i<num2; i++) W[ali2->sids[i]]=1.0f;
      for (i=0; i<num3; i++) W[ali3->sids[i]]=1.0f;
    }
  else
    {
      /* tree-based weighting */
      /* calculate NJ-tree of each alignment */
      tree1=njoin (D, ali1->sids, ali1->num);
      tree2=njoin (D, ali2->sids, ali2->num);
      tree3=njoin (D, ali3->sids, ali3->num);

      /* setup global root */
      root=NODEmake (-2);

      /* create edges between trees and global root */
      edge1=EDGEmake ();
      edge2=EDGEmake ();
      edge3=EDGEmake ();

      /* calculate edge lengths */
      /* here the median problem is needed to be solved for 3 nodes */
      l1=1.0;
      l2=1.0;
      l3=1.0;

#ifdef DEBUG
      fprintf (debugfile, "calcWeights (inserted edge lengths)\n");
      fprintf (debugfile, "l1: %6.3f:\n", l1);
      fprintf (debugfile, "l2: %6.3f:\n", l2);
      fprintf (debugfile, "l3: %6.3f:\n", l3);
      fprintf (debugfile, "\n");
#endif
     
      /* set edge lengths */
      edge1->length=l1;
      edge2->length=l2;
      edge3->length=l3;

      /* insert edges between global root and single trees */
      NODEaddedge (root, tree1, edge1);
      NODEaddedge (root, tree2, edge2);
      NODEaddedge (root, tree3, edge3);

      /* calculate node weights of the entire tree and store them to vector */
      calcNodeWeights (root, W);

      /* clean up tree */
      TREEfree (root);
    }

#ifdef DEBUG
  fprintf (debugfile, "calcWeights (raw weights)\n");
  for (i=0; i<num1; i++) fprintf (debugfile, "%d: %4.2f ", ali1->sids[i], W[ali1->sids[i]]);
  for (i=0; i<num2; i++) fprintf (debugfile, "%d: %4.2f ", ali2->sids[i], W[ali2->sids[i]]);
  for (i=0; i<num3; i++) fprintf (debugfile, "%d: %4.2f ", ali3->sids[i], W[ali3->sids[i]]);
  fprintf (debugfile, "\n");
#endif

  /* normalize weights */
  sum=0.0f;
  for (i=0; i<num; i++) sum+=W[i];
  if (sum>0.0f) for (i=0; i<num; i++) W[i]=(W[i]*(float)num)/sum;
  else for (i=0; i<num; i++) W[i]=1.0f;

#ifdef DEBUG
  fprintf (debugfile, "all sequence weights (normalized)\n");
  for (i=0; i<num; i++) fprintf (debugfile, "id %d: %5.2f\n", i, W[i]);
#endif
}
